# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>
# Written by Matt Prahl <mprahl@redhat.com>

"""Auth system based on the client certificate and FAS account"""
import json
import os
from socket import gethostname
import ssl

import requests
import kerberos
from flask import Response
# Starting with Flask 0.9, the _app_ctx_stack is the correct one,
# before that we need to use the _request_ctx_stack.
try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack
from werkzeug.exceptions import Unauthorized as FlaskUnauthorized
from dogpile.cache import make_region

from module_build_service.errors import Unauthorized, Forbidden
from module_build_service import app, log, conf

try:
    import ldap3
except ImportError:
    log.warn("ldap3 import not found.  ldap/krb disabled.")


client_secrets = None
region = make_region().configure('dogpile.cache.memory')


def _json_loads(content):
    if not isinstance(content, str):
        content = content.decode('utf-8')
    return json.loads(content)


def _load_secrets():
    global client_secrets
    if client_secrets:
        return

    if "OIDC_CLIENT_SECRETS" not in app.config:
        raise Forbidden("OIDC_CLIENT_SECRETS must be set in server config.")

    secrets = _json_loads(open(app.config['OIDC_CLIENT_SECRETS'],
                               'r').read())
    client_secrets = list(secrets.values())[0]


def _get_token_info(token):
    """
    Asks the token_introspection_uri for the validity of a token.
    """
    if not client_secrets:
        return None

    request = {'token': token,
               'token_type_hint': 'Bearer',
               'client_id': client_secrets['client_id'],
               'client_secret': client_secrets['client_secret']}
    headers = {'Content-type': 'application/x-www-form-urlencoded'}

    resp = requests.post(client_secrets['token_introspection_uri'], data=request, headers=headers)
    return resp.json()


def _get_user_info(token):
    """
    Asks the userinfo_uri for more information on a user.
    """
    if not client_secrets:
        return None

    headers = {'authorization': 'Bearer ' + token}
    resp = requests.get(client_secrets['userinfo_uri'], headers=headers)
    return resp.json()


def get_user_oidc(request):
    """
    Returns the client's username and groups based on the OIDC token provided.
    """
    _load_secrets()

    if "authorization" not in request.headers:
        raise Unauthorized("No 'authorization' header found.")

    header = request.headers['authorization'].strip()
    prefix = 'Bearer '
    if not header.startswith(prefix):
        raise Unauthorized("Authorization headers must start with %r" % prefix)

    token = header[len(prefix):].strip()
    try:
        data = _get_token_info(token)
    except Exception as e:
        error = "Cannot verify OIDC token: %s" % str(e)
        log.exception(error)
        raise Exception(error)

    if not data or "active" not in data or not data["active"]:
        raise Unauthorized("OIDC token invalid or expired.")

    if "OIDC_REQUIRED_SCOPE" not in app.config:
        raise Forbidden("OIDC_REQUIRED_SCOPE must be set in server config.")

    presented_scopes = data['scope'].split(' ')
    required_scopes = [
        'openid',
        'https://id.fedoraproject.org/scope/groups',
        app.config["OIDC_REQUIRED_SCOPE"],
    ]
    for scope in required_scopes:
        if scope not in presented_scopes:
            raise Unauthorized("Required OIDC scope %r not present: %r" % (
                scope, presented_scopes))

    try:
        extended_data = _get_user_info(token)
    except Exception as e:
        error = "Cannot verify determine user groups:  %s" % str(e)
        log.exception(error)
        raise Exception(error)

    try:
        groups = set(extended_data['groups'])
    except Exception as e:
        error = "Could not find groups in UserInfo from OIDC %s" % str(e)
        log.exception(extended_data)
        raise Exception(error)

    return data["username"], groups


# Insired by https://pagure.io/waiverdb/blob/master/f/waiverdb/auth.py which was
# inspired by https://github.com/mkomitee/flask-kerberos/blob/master/flask_kerberos.py
class KerberosAuthenticate(object):
    def __init__(self):
        if conf.kerberos_http_host:
            hostname = conf.kerberos_http_host
        else:
            hostname = gethostname()
        self.service_name = "HTTP@{0}".format(hostname)

        # If the config specifies a keytab to use, then override the KRB5_KTNAME
        # environment variable
        if conf.kerberos_keytab:
            os.environ['KRB5_KTNAME'] = conf.kerberos_keytab

        if 'KRB5_KTNAME' in os.environ:
            try:
                principal = kerberos.getServerPrincipalDetails('HTTP', hostname)
            except kerberos.KrbError as error:
                raise Unauthorized(
                    'Kerberos: authentication failed with "{0}"'.format(str(error)))

            log.debug('Kerberos: server is identifying as "{0}"'.format(principal))
        else:
            raise Unauthorized('Kerberos: set the config value of "KERBEROS_KEYTAB" or the '
                               'environment variable "KRB5_KTNAME" to your keytab file')

    def _gssapi_authenticate(self, token):
        """
        Performs GSSAPI Negotiate Authentication
        On success also stashes the server response token for mutual authentication
        at the top of request context with the name kerberos_token, along with the
        authenticated user principal with the name kerberos_user.
        """
        state = None
        ctx = stack.top
        try:
            rc, state = kerberos.authGSSServerInit(self.service_name)
            if rc != kerberos.AUTH_GSS_COMPLETE:
                log.error('Kerberos: unable to initialize server context')
                return None

            rc = kerberos.authGSSServerStep(state, token)
            if rc == kerberos.AUTH_GSS_COMPLETE:
                log.debug('Kerberos: completed GSSAPI negotiation')
                ctx.kerberos_token = kerberos.authGSSServerResponse(state)
                ctx.kerberos_user = kerberos.authGSSServerUserName(state)
                return rc
            elif rc == kerberos.AUTH_GSS_CONTINUE:
                log.debug('Kerberos: continuing GSSAPI negotiation')
                return kerberos.AUTH_GSS_CONTINUE
            else:
                log.debug('Kerberos: unable to step server context')
                return None
        except kerberos.GSSError as error:
            log.error('Kerberos: unable to authenticate: {0}'.format(str(error)))
            return None
        finally:
            if state:
                kerberos.authGSSServerClean(state)

    def process_request(self, token):
        """
        Authenticates the current request using Kerberos.
        """
        kerberos_user = None
        kerberos_token = None
        ctx = stack.top
        rc = self._gssapi_authenticate(token)
        if rc == kerberos.AUTH_GSS_COMPLETE:
            kerberos_user = ctx.kerberos_user
            kerberos_token = ctx.kerberos_token
        elif rc != kerberos.AUTH_GSS_CONTINUE:
            raise Forbidden('Invalid Kerberos ticket')

        return kerberos_user, kerberos_token


def get_user_kerberos(request):
    user = None
    if 'Authorization' not in request.headers:
        response = Response('Unauthorized', 401, {'WWW-Authenticate': 'Negotiate'})
        raise FlaskUnauthorized(response=response)
    header = request.headers.get('Authorization')
    token = ''.join(header.strip().split()[1:])
    user, kerberos_token = KerberosAuthenticate().process_request(token)
    # Remove the realm
    user = user.split('@')[0]
    groups = get_ldap_group_membership(user)
    return user, set(groups)


@region.cache_on_arguments()
def get_ldap_group_membership(uid):
    """ Small wrapper on getting the group membership so that we can use caching
    :param uid: a string of the uid of the user
    :return: a list of groups the user is a member of
    """
    ldap_con = Ldap()
    return ldap_con.get_user_membership(uid)


class Ldap(object):
    """ A class that handles LDAP connections and queries
    """
    connection = None
    base_dn = None

    def __init__(self):
        if not conf.ldap_uri:
            raise Forbidden('LDAP_URI must be set in server config.')
        if conf.ldap_groups_dn:
            self.base_dn = conf.ldap_groups_dn
        else:
            raise Forbidden('LDAP_GROUPS_DN must be set in server config.')

        if conf.ldap_uri.startswith('ldaps://'):
            tls = ldap3.Tls(ca_certs_file='/etc/pki/tls/certs/ca-bundle.crt',
                            validate=ssl.CERT_REQUIRED)
            server = ldap3.Server(conf.ldap_uri, use_ssl=True, tls=tls)
        else:
            server = ldap3.Server(conf.ldap_uri)
        self.connection = ldap3.Connection(server)
        try:
            self.connection.open()
        except ldap3.core.exceptions.LDAPSocketOpenError as error:
            log.error('The connection to "{0}" failed. The following error was raised: {1}'
                      .format(conf.ldap_uri, str(error)))
            raise Forbidden('The connection to the LDAP server failed. Group membership '
                            'couldn\'t be obtained.')

    def get_user_membership(self, uid):
        """ Gets the group membership of a user
        :param uid: a string of the uid of the user
        :return: a list of common names of the posixGroups the user is a member of
        """
        ldap_filter = '(memberUid={0})'.format(uid)
        # Only get the groups in the base container/OU
        self.connection.search(self.base_dn, ldap_filter, search_scope=ldap3.LEVEL,
                               attributes=['cn'])
        groups = self.connection.response
        try:
            return [group['attributes']['cn'][0] for group in groups]
        except KeyError:
            log.exception('The LDAP groups could not be determined based on the search results '
                          'of "{0}"'.format(str(groups)))
            return []


def get_user(request):
    """ Authenticates the user and returns the username and group name
    :param request: a Flask request
    :return: a tuple with a string representing the user name and a set with the user's group
    membership such as ('mprahl', {'factory2', 'devel'})
    """
    if conf.no_auth is True:
        log.debug('Authorization is disabled.')
        return 'anonymous', {'packager'}

    get_user_func_name = 'get_user_{0}'.format(conf.auth_method)
    get_user_func = globals().get(get_user_func_name)
    if not get_user_func:
        raise RuntimeError('The function "{0}" is not implemented'.format(get_user_func_name))
    return get_user_func(request)



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Matt Prahl <mprahl@redhat.com

import os
from datetime import datetime, timedelta
from mock import patch
import time
import hashlib
from traceback import extract_stack

import koji
import module_build_service
from module_build_service import db
from module_build_service.utils import get_rpm_release, import_mmd, load_mmd
from module_build_service.config import init_config
from module_build_service.models import ModuleBuild, ComponentBuild, make_session, BUILD_STATES
from module_build_service import glib, Modulemd


base_dir = os.path.dirname(__file__)
app = module_build_service.app
conf = init_config(app)


def patch_config():
    # add test builders for all resolvers
    with_test_builders = dict()
    for k, v in module_build_service.config.SUPPORTED_RESOLVERS.items():
        v['builders'].extend(['test', 'testlocal'])
        with_test_builders[k] = v
    patch("module_build_service.config.SUPPORTED_RESOLVERS", with_test_builders)


patch_config()


def patch_zeromq_time_sleep():
    """
    We use moksha.hub in some tests. We used dummy zerombq backend which
    connects to /dev/null, but zeromq.py contains time.sleep(1) to ensure
    that sockets are listening properly. This is not needed for our dummy
    use-case and it slows down tests.

    This method patches time.sleep called from "zeromq.py" file to be noop,
    but calls the real time.sleep otherwise.
    """
    global _orig_time_sleep
    _orig_time_sleep = time.sleep

    def mocked_time_sleep(n):
        global _orig_time_sleep
        if n == 1:
            tb = extract_stack()
            try:
                if tb[-4][0].endswith("zeromq.py"):
                    return
            except IndexError:
                pass
        _orig_time_sleep(n)

    ts = patch("time.sleep").start()
    ts.side_effect = mocked_time_sleep


patch_zeromq_time_sleep()


def clean_database(add_platform_module=True):
    db.session.commit()
    db.drop_all()
    db.create_all()
    if add_platform_module:
        mmd = load_mmd(os.path.join(base_dir, 'staged_data', 'platform.yaml'), True)
        import_mmd(db.session, mmd)


def init_data(data_size=10, contexts=False):
    """
    Creates data_size * 3 modules in database in different states and
    with different component builds. See _populate_data for more info.

    :param bool contexts: If True, multiple streams and contexts in each stream
        are generated for 'nginx' module.
    """
    clean_database()
    with make_session(conf) as session:
        _populate_data(session, data_size, contexts=contexts)


def _populate_data(session, data_size=10, contexts=False):
    num_contexts = 2 if contexts else 1
    for index in range(data_size):
        for context in range(num_contexts):
            build_one = ModuleBuild()
            build_one.name = 'nginx'
            build_one.stream = '1'
            build_one.version = 2 + index
            build_one.state = BUILD_STATES['ready']
            if contexts:
                build_one.stream = str(index)
                unique_hash = hashlib.sha1("%s:%s:%d:%d" % (
                    build_one.name, build_one.stream, build_one.version, context)).hexdigest()
                build_one.build_context = unique_hash
                build_one.runtime_context = unique_hash
                build_one.ref_build_context = unique_hash
                combined_hashes = '{0}:{1}'.format(unique_hash, unique_hash)
                build_one.context = hashlib.sha1(combined_hashes).hexdigest()[:8]
            with open(os.path.join(base_dir, "staged_data", "nginx_mmd.yaml")) as mmd:
                build_one.modulemd = mmd.read()
            build_one.koji_tag = 'module-nginx-1.2'
            build_one.scmurl = ('git://pkgs.domain.local/modules/nginx?'
                                '#ba95886c7a443b36a9ce31abda1f9bef22f2f8c9')
            build_one.batch = 2
            # https://www.youtube.com/watch?v=iQGwrK_yDEg
            build_one.owner = 'Moe Szyslak'
            build_one.time_submitted = \
                datetime(2016, 9, 3, 11, 23, 20) + timedelta(minutes=(index * 10))
            build_one.time_modified = \
                datetime(2016, 9, 3, 11, 25, 32) + timedelta(minutes=(index * 10))
            build_one.time_completed = \
                datetime(2016, 9, 3, 11, 25, 32) + timedelta(minutes=(index * 10))
            build_one.rebuild_strategy = 'changed-and-after'
            session.add(build_one)
            session.commit()
            build_one_component_release = get_rpm_release(build_one)

            component_one_build_one = ComponentBuild()
            component_one_build_one.package = 'nginx'
            component_one_build_one.scmurl = \
                ('git://pkgs.domain.local/rpms/nginx?'
                 '#ga95886c8a443b36a9ce31abda1f9bed22f2f8c3')
            component_one_build_one.format = 'rpms'
            component_one_build_one.task_id = 12312345 + index
            component_one_build_one.state = koji.BUILD_STATES['COMPLETE']
            component_one_build_one.nvr = 'nginx-1.10.1-2.{0}'.format(build_one_component_release)
            component_one_build_one.batch = 1
            component_one_build_one.module_id = 2 + index * 3
            component_one_build_one.tagged = True
            component_one_build_one.tagged_in_final = True

            component_two_build_one = ComponentBuild()
            component_two_build_one.package = 'module-build-macros'
            component_two_build_one.scmurl = \
                ('/tmp/module_build_service-build-macrosWZUPeK/SRPMS/'
                 'module-build-macros-0.1-1.module_nginx_1_2.src.rpm')
            component_two_build_one.format = 'rpms'
            component_two_build_one.task_id = 12312321 + index
            component_two_build_one.state = koji.BUILD_STATES['COMPLETE']
            component_two_build_one.nvr = \
                'module-build-macros-01-1.{0}'.format(build_one_component_release)
            component_two_build_one.batch = 2
            component_two_build_one.module_id = 2 + index * 3
            component_two_build_one.tagged = True
            component_two_build_one.tagged_in_final = True

        build_two = ModuleBuild()
        build_two.name = 'postgressql'
        build_two.stream = '1'
        build_two.version = 2 + index
        build_two.state = BUILD_STATES['done']
        build_two.modulemd = ''  # Skipping since no tests rely on it
        build_two.koji_tag = 'module-postgressql-1.2'
        build_two.scmurl = ('git://pkgs.domain.local/modules/postgressql?'
                            '#aa95886c7a443b36a9ce31abda1f9bef22f2f8c9')
        build_two.batch = 2
        build_two.owner = 'some_user'
        build_two.time_submitted = \
            datetime(2016, 9, 3, 12, 25, 33) + timedelta(minutes=(index * 10))
        build_two.time_modified = \
            datetime(2016, 9, 3, 12, 27, 19) + timedelta(minutes=(index * 10))
        build_two.time_completed = \
            datetime(2016, 9, 3, 11, 27, 19) + timedelta(minutes=(index * 10))
        build_two.rebuild_strategy = 'changed-and-after'
        session.add(build_two)
        session.commit()
        build_two_component_release = get_rpm_release(build_two)

        component_one_build_two = ComponentBuild()
        component_one_build_two.package = 'postgresql'
        component_one_build_two.scmurl = \
            ('git://pkgs.domain.local/rpms/postgresql?'
             '#dc95586c4a443b26a9ce38abda1f9bed22f2f8c3')
        component_one_build_two.format = 'rpms'
        component_one_build_two.task_id = 2433433 + index
        component_one_build_two.state = koji.BUILD_STATES['COMPLETE']
        component_one_build_two.nvr = 'postgresql-9.5.3-4.{0}'.format(build_two_component_release)
        component_one_build_two.batch = 2
        component_one_build_two.module_id = 3 + index * 3
        component_one_build_two.tagged = True
        component_one_build_two.tagged_in_final = True

        component_two_build_two = ComponentBuild()
        component_two_build_two.package = 'module-build-macros'
        component_two_build_two.scmurl = \
            ('/tmp/module_build_service-build-macrosWZUPeK/SRPMS/'
             'module-build-macros-0.1-1.module_postgresql_1_2.src.rpm')
        component_two_build_two.format = 'rpms'
        component_two_build_two.task_id = 47383993 + index
        component_two_build_two.state = koji.BUILD_STATES['COMPLETE']
        component_two_build_two.nvr = \
            'module-build-macros-01-1.{0}'.format(build_two_component_release)
        component_two_build_two.batch = 1
        component_two_build_two.module_id = 3 + index * 3
        component_one_build_two.tagged = True
        component_one_build_two.build_time_only = True

        build_three = ModuleBuild()
        build_three.name = 'testmodule'
        build_three.stream = '4.3.43'
        build_three.version = 6 + index
        build_three.state = BUILD_STATES['wait']
        build_three.modulemd = ''  # Skipping because no tests rely on it
        build_three.koji_tag = None
        build_three.scmurl = ('git://pkgs.domain.local/modules/testmodule?'
                              '#ca95886c7a443b36a9ce31abda1f9bef22f2f8c9')
        build_three.batch = 0
        build_three.owner = 'some_other_user'
        build_three.time_submitted = \
            datetime(2016, 9, 3, 12, 28, 33) + timedelta(minutes=(index * 10))
        build_three.time_modified = \
            datetime(2016, 9, 3, 12, 28, 40) + timedelta(minutes=(index * 10))
        build_three.time_completed = None
        build_three.rebuild_strategy = 'changed-and-after'
        session.add(build_three)
        session.commit()
        build_three_component_release = get_rpm_release(build_three)

        component_one_build_three = ComponentBuild()
        component_one_build_three.package = 'rubygem-rails'
        component_one_build_three.scmurl = \
            ('git://pkgs.domain.local/rpms/rubygem-rails?'
             '#dd55886c4a443b26a9ce38abda1f9bed22f2f8c3')
        component_one_build_three.format = 'rpms'
        component_one_build_three.task_id = 2433433 + index
        component_one_build_three.state = koji.BUILD_STATES['FAILED']
        component_one_build_three.nvr = \
            'postgresql-9.5.3-4.{0}'.format(build_three_component_release)
        component_one_build_three.batch = 2
        component_one_build_three.module_id = 4 + index * 3

        component_two_build_three = ComponentBuild()
        component_two_build_three.package = 'module-build-macros'
        component_two_build_three.scmurl = \
            ('/tmp/module_build_service-build-macrosWZUPeK/SRPMS/'
             'module-build-macros-0.1-1.module_testmodule_1_2.src.rpm')
        component_two_build_three.format = 'rpms'
        component_two_build_three.task_id = 47383993 + index
        component_two_build_three.state = koji.BUILD_STATES['COMPLETE']
        component_two_build_three.nvr = \
            'module-build-macros-01-1.{0}'.format(build_three_component_release)
        component_two_build_three.batch = 1
        component_two_build_three.module_id = 4 + index * 3
        component_two_build_three.tagged = True
        component_two_build_three.build_time_only = True

        session.add(component_one_build_one)
        session.add(component_two_build_one)
        session.add(component_one_build_two)
        session.add(component_two_build_two)
        session.add(component_one_build_three)
        session.add(component_two_build_three)
        session.commit()


def scheduler_init_data(tangerine_state=None):
    """ Creates a testmodule in the building state with all the components in the same batch
    """
    clean_database()

    current_dir = os.path.dirname(__file__)
    formatted_testmodule_yml_path = os.path.join(
        current_dir, 'staged_data', 'formatted_testmodule.yaml')
    mmd = Modulemd.Module().new_from_file(formatted_testmodule_yml_path)
    mmd.upgrade()
    mmd.get_rpm_components()['tangerine'].set_buildorder(0)

    build_one = module_build_service.models.ModuleBuild()
    build_one.name = 'testmodule'
    build_one.stream = 'master'
    build_one.version = 20170109091357
    build_one.state = BUILD_STATES['build']
    build_one.build_context = 'ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0'
    build_one.runtime_context = 'ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0'
    build_one.context = '7c29193d'
    build_one.koji_tag = 'module-testmodule-master-20170109091357-7c29193d'
    build_one.scmurl = 'git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#ff1ea79'
    if tangerine_state:
        build_one.batch = 3
    else:
        build_one.batch = 2
    # https://www.youtube.com/watch?v=iOKymYVSaJE
    build_one.owner = 'Buzz Lightyear'
    build_one.time_submitted = datetime(2017, 2, 15, 16, 8, 18)
    build_one.time_modified = datetime(2017, 2, 15, 16, 19, 35)
    build_one.rebuild_strategy = 'changed-and-after'
    build_one.modulemd = mmd.dumps()
    build_one_component_release = get_rpm_release(build_one)

    component_one_build_one = module_build_service.models.ComponentBuild()
    component_one_build_one.package = 'perl-Tangerine'
    component_one_build_one.scmurl = \
        ('git://pkgs.fedoraproject.org/rpms/perl-Tangerine'
         '?#4ceea43add2366d8b8c5a622a2fb563b625b9abf')
    component_one_build_one.format = 'rpms'
    component_one_build_one.task_id = 90276227
    component_one_build_one.state = koji.BUILD_STATES['COMPLETE']
    component_one_build_one.nvr = \
        'perl-Tangerine-0.23-1.{0}'.format(build_one_component_release)
    component_one_build_one.batch = 2
    component_one_build_one.module_id = 2
    component_one_build_one.ref = '4ceea43add2366d8b8c5a622a2fb563b625b9abf'
    component_one_build_one.tagged = True
    component_one_build_one.tagged_in_final = True

    component_two_build_one = module_build_service.models.ComponentBuild()
    component_two_build_one.package = 'perl-List-Compare'
    component_two_build_one.scmurl = \
        ('git://pkgs.fedoraproject.org/rpms/perl-List-Compare'
         '?#76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb')
    component_two_build_one.format = 'rpms'
    component_two_build_one.task_id = 90276228
    component_two_build_one.state = koji.BUILD_STATES['COMPLETE']
    component_two_build_one.nvr = \
        'perl-List-Compare-0.53-5.{0}'.format(build_one_component_release)
    component_two_build_one.batch = 2
    component_two_build_one.module_id = 2
    component_two_build_one.ref = '76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb'
    component_two_build_one.tagged = True
    component_two_build_one.tagged_in_final = True

    component_three_build_one = module_build_service.models.ComponentBuild()
    component_three_build_one.package = 'tangerine'
    component_three_build_one.scmurl = \
        ('git://pkgs.fedoraproject.org/rpms/tangerine'
         '?#fbed359411a1baa08d4a88e0d12d426fbf8f602c')
    component_three_build_one.format = 'rpms'
    component_three_build_one.batch = 3
    component_three_build_one.module_id = 2
    component_three_build_one.ref = 'fbed359411a1baa08d4a88e0d12d426fbf8f602c'
    component_three_build_one.state = tangerine_state
    if tangerine_state:
        component_three_build_one.task_id = 90276315
        component_three_build_one.nvr = \
            'tangerine-0.22-3.{0}'.format(build_one_component_release)
    if tangerine_state == koji.BUILD_STATES['COMPLETE']:
        component_three_build_one.tagged = True
        component_three_build_one.tagged_in_final = True

    component_four_build_one = module_build_service.models.ComponentBuild()
    component_four_build_one.package = 'module-build-macros'
    component_four_build_one.scmurl = \
        ('/tmp/module_build_service-build-macrosqr4AWH/SRPMS/module-build-'
         'macros-0.1-1.module_testmodule_master_20170109091357.src.rpm')
    component_four_build_one.format = 'rpms'
    component_four_build_one.task_id = 90276181
    component_four_build_one.state = koji.BUILD_STATES['COMPLETE']
    component_four_build_one.nvr = \
        'module-build-macros-0.1-1.{0}'.format(build_one_component_release)
    component_four_build_one.batch = 1
    component_four_build_one.module_id = 2
    component_four_build_one.tagged = True
    component_four_build_one.build_time_only = True

    with make_session(conf) as session:
        session.add(build_one)
        session.add(component_one_build_one)
        session.add(component_two_build_one)
        session.add(component_three_build_one)
        session.add(component_four_build_one)
        session.commit()


def reuse_component_init_data():
    clean_database()

    current_dir = os.path.dirname(__file__)
    formatted_testmodule_yml_path = os.path.join(
        current_dir, 'staged_data', 'formatted_testmodule.yaml')
    mmd = Modulemd.Module().new_from_file(formatted_testmodule_yml_path)
    mmd.upgrade()

    build_one = module_build_service.models.ModuleBuild()
    build_one.name = 'testmodule'
    build_one.stream = 'master'
    build_one.version = 20170109091357
    build_one.state = BUILD_STATES['ready']
    build_one.ref_build_context = 'ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0'
    build_one.runtime_context = 'ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0'
    build_one.build_context = 'ac4de1c346dcf09ce77d38cd4e75094ec1c08eb1'
    build_one.context = '78e4a6fd'
    build_one.koji_tag = 'module-testmodule-master-20170109091357-78e4a6fd'
    build_one.scmurl = 'git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#ff1ea79'
    build_one.batch = 3
    build_one.owner = 'Tom Brady'
    build_one.time_submitted = datetime(2017, 2, 15, 16, 8, 18)
    build_one.time_modified = datetime(2017, 2, 15, 16, 19, 35)
    build_one.time_completed = datetime(2017, 2, 15, 16, 19, 35)
    build_one.rebuild_strategy = 'changed-and-after'
    build_one_component_release = get_rpm_release(build_one)
    mmd.set_version(build_one.version)
    xmd = glib.from_variant_dict(mmd.get_xmd())
    xmd['mbs']['scmurl'] = build_one.scmurl
    xmd['mbs']['commit'] = 'ff1ea79fc952143efeed1851aa0aa006559239ba'
    mmd.set_xmd(glib.dict_values(xmd))
    build_one.modulemd = mmd.dumps()

    component_one_build_one = module_build_service.models.ComponentBuild()
    component_one_build_one.package = 'perl-Tangerine'
    component_one_build_one.scmurl = \
        ('git://pkgs.fedoraproject.org/rpms/perl-Tangerine'
         '?#4ceea43add2366d8b8c5a622a2fb563b625b9abf')
    component_one_build_one.format = 'rpms'
    component_one_build_one.task_id = 90276227
    component_one_build_one.state = koji.BUILD_STATES['COMPLETE']
    component_one_build_one.nvr = \
        'perl-Tangerine-0.23-1.{0}'.format(build_one_component_release)
    component_one_build_one.batch = 2
    component_one_build_one.module_id = 2
    component_one_build_one.ref = '4ceea43add2366d8b8c5a622a2fb563b625b9abf'
    component_one_build_one.tagged = True
    component_one_build_one.tagged_in_final = True
    component_two_build_one = module_build_service.models.ComponentBuild()
    component_two_build_one.package = 'perl-List-Compare'
    component_two_build_one.scmurl = \
        ('git://pkgs.fedoraproject.org/rpms/perl-List-Compare'
         '?#76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb')
    component_two_build_one.format = 'rpms'
    component_two_build_one.task_id = 90276228
    component_two_build_one.state = koji.BUILD_STATES['COMPLETE']
    component_two_build_one.nvr = \
        'perl-List-Compare-0.53-5.{0}'.format(build_one_component_release)
    component_two_build_one.batch = 2
    component_two_build_one.module_id = 2
    component_two_build_one.ref = '76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb'
    component_two_build_one.tagged = True
    component_two_build_one.tagged_in_final = True
    component_three_build_one = module_build_service.models.ComponentBuild()
    component_three_build_one.package = 'tangerine'
    component_three_build_one.scmurl = \
        ('git://pkgs.fedoraproject.org/rpms/tangerine'
         '?#fbed359411a1baa08d4a88e0d12d426fbf8f602c')
    component_three_build_one.format = 'rpms'
    component_three_build_one.task_id = 90276315
    component_three_build_one.state = koji.BUILD_STATES['COMPLETE']
    component_three_build_one.nvr = \
        'tangerine-0.22-3.{0}'.format(build_one_component_release)
    component_three_build_one.batch = 3
    component_three_build_one.module_id = 2
    component_three_build_one.ref = 'fbed359411a1baa08d4a88e0d12d426fbf8f602c'
    component_three_build_one.tagged = True
    component_three_build_one.tagged_in_final = True
    component_four_build_one = module_build_service.models.ComponentBuild()
    component_four_build_one.package = 'module-build-macros'
    component_four_build_one.scmurl = \
        ('/tmp/module_build_service-build-macrosqr4AWH/SRPMS/module-build-'
         'macros-0.1-1.module_testmodule_master_20170109091357.src.rpm')
    component_four_build_one.format = 'rpms'
    component_four_build_one.task_id = 90276181
    component_four_build_one.state = koji.BUILD_STATES['COMPLETE']
    component_four_build_one.nvr = \
        'module-build-macros-0.1-1.{0}'.format(build_one_component_release)
    component_four_build_one.batch = 1
    component_four_build_one.module_id = 2
    component_four_build_one.tagged = True
    component_four_build_one.build_time_only = True

    build_two = module_build_service.models.ModuleBuild()
    build_two.name = 'testmodule'
    build_two.stream = 'master'
    build_two.version = 20170219191323
    build_two.state = BUILD_STATES['build']
    build_two.ref_build_context = 'ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0'
    build_two.runtime_context = 'ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0'
    build_two.build_context = 'ac4de1c346dcf09ce77d38cd4e75094ec1c08eb1'
    build_two.context = 'c40c156c'
    build_two.koji_tag = 'module-testmodule-master-20170219191323-c40c156c'
    build_two.scmurl = 'git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#55f4a0a'
    build_two.batch = 1
    build_two.owner = 'Tom Brady'
    build_two.time_submitted = datetime(2017, 2, 19, 16, 8, 18)
    build_two.time_modified = datetime(2017, 2, 19, 16, 8, 18)
    build_two.rebuild_strategy = 'changed-and-after'
    build_two_component_release = get_rpm_release(build_two)
    mmd.set_version(build_one.version)
    xmd = glib.from_variant_dict(mmd.get_xmd())
    xmd['mbs']['scmurl'] = build_one.scmurl
    xmd['mbs']['commit'] = '55f4a0a2e6cc255c88712a905157ab39315b8fd8'
    mmd.set_xmd(glib.dict_values(xmd))
    build_two.modulemd = mmd.dumps()

    component_one_build_two = module_build_service.models.ComponentBuild()
    component_one_build_two.package = 'perl-Tangerine'
    component_one_build_two.scmurl = \
        ('git://pkgs.fedoraproject.org/rpms/perl-Tangerine'
         '?#4ceea43add2366d8b8c5a622a2fb563b625b9abf')
    component_one_build_two.format = 'rpms'
    component_one_build_two.batch = 2
    component_one_build_two.module_id = 3
    component_one_build_two.ref = '4ceea43add2366d8b8c5a622a2fb563b625b9abf'
    component_two_build_two = module_build_service.models.ComponentBuild()
    component_two_build_two.package = 'perl-List-Compare'
    component_two_build_two.scmurl = \
        ('git://pkgs.fedoraproject.org/rpms/perl-List-Compare'
         '?#76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb')
    component_two_build_two.format = 'rpms'
    component_two_build_two.batch = 2
    component_two_build_two.module_id = 3
    component_two_build_two.ref = '76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb'
    component_three_build_two = module_build_service.models.ComponentBuild()
    component_three_build_two.package = 'tangerine'
    component_three_build_two.scmurl = \
        ('git://pkgs.fedoraproject.org/rpms/tangerine'
         '?#fbed359411a1baa08d4a88e0d12d426fbf8f602c')
    component_three_build_two.format = 'rpms'
    component_three_build_two.batch = 3
    component_three_build_two.module_id = 3
    component_three_build_two.ref = 'fbed359411a1baa08d4a88e0d12d426fbf8f602c'
    component_four_build_two = module_build_service.models.ComponentBuild()
    component_four_build_two.package = 'module-build-macros'
    component_four_build_two.scmurl = \
        ('/tmp/module_build_service-build-macrosqr4AWH/SRPMS/module-build-'
         'macros-0.1-1.module_testmodule_master_20170219191323.src.rpm')
    component_four_build_two.format = 'rpms'
    component_four_build_two.task_id = 90276186
    component_four_build_two.state = koji.BUILD_STATES['COMPLETE']
    component_four_build_two.nvr = \
        'module-build-macros-0.1-1.{0}'.format(build_two_component_release)
    component_four_build_two.batch = 1
    component_four_build_two.module_id = 3
    component_four_build_two.tagged = True
    component_four_build_two.build_time_only = True

    with make_session(conf) as session:
        session.add(build_one)
        session.add(component_one_build_one)
        session.add(component_two_build_one)
        session.add(component_three_build_one)
        session.add(component_four_build_one)
        session.add(build_two)
        session.add(component_one_build_two)
        session.add(component_two_build_two)
        session.add(component_three_build_two)
        session.add(component_four_build_two)
        session.commit()


def reuse_shared_userspace_init_data():
    clean_database()

    with make_session(conf) as session:
        # Create shared-userspace-570, state is COMPLETE, all components
        # are properly built.
        current_dir = os.path.dirname(__file__)
        formatted_testmodule_yml_path = os.path.join(
            current_dir, 'staged_data', 'shared-userspace-570.yaml')
        mmd = Modulemd.Module().new_from_file(formatted_testmodule_yml_path)
        mmd.upgrade()

        build_one = module_build_service.models.ModuleBuild()
        build_one.name = mmd.get_name()
        build_one.stream = mmd.get_stream()
        build_one.version = mmd.get_version()
        build_one.build_context = 'e046b867a400a06a3571f3c71142d497895fefbe'
        build_one.runtime_context = '50dd3eb5dde600d072e45d4120e1548ce66bc94a'
        build_one.state = BUILD_STATES['ready']
        build_one.modulemd = mmd.dumps()
        build_one.koji_tag = 'module-shared-userspace-f26-20170601141014-75f92abb'
        build_one.scmurl = ('git://pkgs.stg.fedoraproject.org/modules/testmodule.'
                            'git?#7fea453')
        build_one.batch = 16
        build_one.owner = 'Tom Brady'
        build_one.time_submitted = datetime(2017, 2, 15, 16, 8, 18)
        build_one.time_modified = datetime(2017, 2, 15, 16, 19, 35)
        build_one.time_completed = datetime(2017, 2, 15, 16, 19, 35)
        build_one.rebuild_strategy = 'changed-and-after'

        session.add(build_one)

        components = mmd.get_rpm_components().values()
        components.sort(key=lambda x: x.get_buildorder())
        previous_buildorder = None
        batch = 1
        for pkg in components:
            # Increment the batch number when buildorder increases.
            if previous_buildorder != pkg.get_buildorder():
                previous_buildorder = pkg.get_buildorder()
                batch += 1

            pkgref = mmd.get_xmd()['mbs']['rpms'][pkg.get_name()]['ref']
            full_url = pkg.get_repository() + "?#" + pkgref
            build = module_build_service.models.ComponentBuild(
                module_id=2,
                package=pkg.get_name(),
                format="rpms",
                scmurl=full_url,
                batch=batch,
                ref=pkgref,
                state=1,
                tagged=True,
                tagged_in_final=True
            )
            session.add(build)

        # Create shared-userspace-577, state is WAIT, no component built
        formatted_testmodule_yml_path = os.path.join(
            current_dir, 'staged_data', 'shared-userspace-577.yaml')
        mmd2 = Modulemd.Module().new_from_file(formatted_testmodule_yml_path)
        mmd2.upgrade()

        build_one = module_build_service.models.ModuleBuild()
        build_one.name = mmd2.get_name()
        build_one.stream = mmd2.get_stream()
        build_one.version = mmd2.get_version()
        build_one.build_context = 'e046b867a400a06a3571f3c71142d497895fefbe'
        build_one.runtime_context = '50dd3eb5dde600d072e45d4120e1548ce66bc94a'
        build_one.state = BUILD_STATES['done']
        build_one.modulemd = mmd2.dumps()
        build_one.koji_tag = 'module-shared-userspace-f26-20170605091544-75f92abb'
        build_one.scmurl = ('git://pkgs.stg.fedoraproject.org/modules/testmodule.'
                            'git?#7fea453')
        build_one.batch = 0
        build_one.owner = 'Tom Brady'
        build_one.time_submitted = datetime(2017, 2, 15, 16, 8, 18)
        build_one.time_modified = datetime(2017, 2, 15, 16, 19, 35)
        build_one.time_completed = datetime(2017, 2, 15, 16, 19, 35)
        build_one.rebuild_strategy = 'changed-and-after'

        session.add(build_one)

        components2 = mmd2.get_rpm_components().values()
        # Store components to database in different order than for 570 to
        # reproduce the reusing issue.
        components2.sort(key=lambda x: len(x.get_name()))
        components2.sort(key=lambda x: x.get_buildorder())
        previous_buildorder = None
        batch = 1
        for pkg in components2:
            # Increment the batch number when buildorder increases.
            if previous_buildorder != pkg.get_buildorder():
                previous_buildorder = pkg.get_buildorder()
                batch += 1

            pkgref = mmd2.get_xmd()['mbs']['rpms'][pkg.get_name()]['ref']
            full_url = pkg.get_repository() + "?#" + pkgref
            build = module_build_service.models.ComponentBuild(
                module_id=3,
                package=pkg.get_name(),
                format="rpms",
                scmurl=full_url,
                batch=batch,
                ref=pkgref
            )
            session.add(build)



"""Add cg_build_koji_tag

Revision ID: edb537dd1e8c
Revises: c11a3cfec2a9
Create Date: 2017-09-22 13:50:41.433144

"""

# revision identifiers, used by Alembic.
revision = 'edb537dd1e8c'
down_revision = 'c11a3cfec2a9'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('module_builds', sa.Column('cg_build_koji_tag', sa.String(), nullable=True))


def downgrade():
    op.drop_column('module_builds', 'cg_build_koji_tag')



"""Add ModuleBuid.ref_build_context.

Revision ID: caeae7a4f537
Revises: 9ca1c166f426
Create Date: 2018-04-18 13:37:40.365129

"""

# revision identifiers, used by Alembic.
revision = 'caeae7a4f537'
down_revision = '9ca1c166f426'

from alembic import op
import sqlalchemy as sa

# Data migration imports
from module_build_service import Modulemd
import hashlib
import json
from collections import OrderedDict


modulebuild = sa.Table(
    'module_builds',
    sa.MetaData(),
    sa.Column('id', sa.Integer, primary_key=True),
    sa.Column('modulemd', sa.String()),
    sa.Column('build_context', sa.String()),
    sa.Column('runtime_context', sa.String()),
)


def upgrade():
    connection = op.get_bind()
    with op.batch_alter_table('module_builds') as b:
        b.alter_column('build_context', new_column_name='ref_build_context')

    op.add_column('module_builds', sa.Column('build_context', sa.String()))

    # Determine what the contexts should be based on the modulemd
    for build in connection.execute(modulebuild.select()):
        if not build.modulemd:
            continue
        try:
            mmd = Modulemd.Module().new_from_string(build.modulemd)
            mmd.upgrade()
        except Exception:
            # If the modulemd isn't parseable then skip this build
            continue

        mbs_xmd = mmd.get_xmd().get('mbs', {})
        # Skip the non-MSE builds, so the "context" will be set default one
        # in models.ModuleBuild.
        if "mse" not in mbs_xmd.keys() or not mbs_xmd["mse"]:
            continue

        # It's possible this module build was built before MBS filled out xmd or before MBS
        # filled out the requires in xmd. We also have to use keys because GLib.Variant
        # doesn't support `in` directly.
        if 'buildrequires' not in mbs_xmd.keys():
            continue
        # Get the streams of buildrequires and hash it.
        mmd_formatted_buildrequires = {
            dep: info['stream'] for dep, info in mbs_xmd["buildrequires"].items()}
        property_json = json.dumps(OrderedDict(sorted(mmd_formatted_buildrequires.items())))
        context = hashlib.sha1(property_json).hexdigest()

        # Update the database now
        connection.execute(
            modulebuild.update().where(modulebuild.c.id == build.id).values(
                build_context=context))


def downgrade():
    op.drop_column('module_builds', 'build_context')
    with op.batch_alter_table('module_builds') as b:
        b.alter_column('ref_build_context', new_column_name='build_context')



"""Add ModuleBuild.context

Revision ID: c8e2fc555399
Revises: caeae7a4f537
Create Date: 2018-04-26 22:57:11.398121

"""

# revision identifiers, used by Alembic.
revision = 'c8e2fc555399'
down_revision = 'caeae7a4f537'

from alembic import op
import sqlalchemy as sa

import hashlib


modulebuild = sa.Table(
    'module_builds',
    sa.MetaData(),
    sa.Column('id', sa.Integer, primary_key=True),
    sa.Column('modulemd', sa.String()),
    sa.Column('build_context', sa.String()),
    sa.Column('runtime_context', sa.String()),
    sa.Column('context', sa.String()),
)


def upgrade():
    connection = op.get_bind()

    op.add_column('module_builds', sa.Column('context', sa.String(), nullable=False, server_default='00000000'))

    for build in connection.execute(modulebuild.select()):
        if build.build_context and build.runtime_context:
            combined_hashes = '{0}:{1}'.format(
                build.build_context, build.runtime_context).encode('utf-8')
            context = hashlib.sha1(combined_hashes).hexdigest()[:8]
            connection.execute(
                modulebuild.update().where(modulebuild.c.id == build.id).values(
                    context=context))

def downgrade():
    with op.batch_alter_table('module_builds') as b:
        b.drop_column('context')



"""Do nothing.

This used to be an upgrade that inserted a fake base-runtime module, but the
code was removed as a result of
https://pagure.io/fm-orchestrator/pull-request/225

Revision ID: 0ef60c3ed440
Revises: 145347916a56
Create Date: 2016-11-17 15:39:22.984051

"""


# revision identifiers, used by Alembic.
revision = '0ef60c3ed440'
down_revision = '145347916a56'


def upgrade():
    pass

def downgrade():
    pass



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>
#            Matt Prahl <mprahl@redhat.com>
#            Jan Kaluza <jkaluza@redhat.com>
import functools
import inspect
import hashlib
import time
from datetime import datetime

from module_build_service import conf, log, models
from module_build_service.errors import ValidationError, ProgrammingError


def scm_url_schemes(terse=False):
    """
    Definition of URL schemes supported by both frontend and scheduler.

    NOTE: only git URLs in the following formats are supported atm:
        git://
        git+http://
        git+https://
        git+rsync://
        http://
        https://
        file://

    :param terse=False: Whether to return terse list of unique URL schemes
                        even without the "://".
    """

    scm_types = {
        "git": ("git://", "git+http://", "git+https://",
                "git+rsync://", "http://", "https://", "file://")
    }

    if not terse:
        return scm_types
    else:
        scheme_list = []
        for scm_type, scm_schemes in scm_types.items():
            scheme_list.extend([scheme[:-3] for scheme in scm_schemes])
        return list(set(scheme_list))


def retry(timeout=conf.net_timeout, interval=conf.net_retry_interval, wait_on=Exception):
    """ A decorator that allows to retry a section of code...
    ...until success or timeout.
    """
    def wrapper(function):
        @functools.wraps(function)
        def inner(*args, **kwargs):
            start = time.time()
            while True:
                try:
                    return function(*args, **kwargs)
                except wait_on as e:
                    log.warn("Exception %r raised from %r.  Retry in %rs" % (
                        e, function, interval))
                    time.sleep(interval)
                    if (time.time() - start) >= timeout:
                        raise  # This re-raises the last exception.
        return inner
    return wrapper


def module_build_state_from_msg(msg):
    state = int(msg.module_build_state)
    # TODO better handling
    assert state in models.BUILD_STATES.values(), (
        'state=%s(%s) is not in %s'
        % (state, type(state), list(models.BUILD_STATES.values())))
    return state


def generate_koji_tag(name, stream, version, context, max_length=256):
    """
    Generate a koji tag from name, stream, version and context.

    :param name: a module's name
    :param stream: a module's stream
    :param version: a module's version
    :param context: a module's context
    :kwarg max_length: the maximum length the Koji tag can be before falling back to
    the old format of "module-<hash>"
    :return: a Koji tag
    """
    nsvc_list = [name, stream, str(version), context]
    nsvc_tag = 'module-' + '-'.join(nsvc_list)
    if len(nsvc_tag) + len('-build') > max_length:
        # Fallback to the old format of 'module-<hash>' if the generated koji tag
        # name is longer than max_length
        nsvc_hash = hashlib.sha1('.'.join(nsvc_list)).hexdigest()[:16]
        return 'module-' + nsvc_hash
    return nsvc_tag


def validate_koji_tag(tag_arg_names, pre='', post='-', dict_key='name'):
    """
    Used as a decorator validates koji tag arg(s)' value(s)
    against configurable list of koji tag prefixes.
    Supported arg value types are: dict, list, str

    :param tag_arg_names: Str or list of parameters to validate.
    :param pre: Prepend this optional string (e.g. '.' in case of disttag
    validation) to each koji tag prefix.
    :param post: Append this string/delimiter ('-' by default) to each koji
    tag prefix.
    :param dict_key: In case of a dict arg, inspect this key ('name' by default).
    """

    if not isinstance(tag_arg_names, list):
        tag_arg_names = [tag_arg_names]

    def validation_decorator(function):
        def wrapper(*args, **kwargs):
            call_args = inspect.getcallargs(function, *args, **kwargs)

            for tag_arg_name in tag_arg_names:
                err_subject = "Koji tag validation:"

                # If any of them don't appear in the function, then fail.
                if tag_arg_name not in call_args:
                    raise ProgrammingError(
                        '{} Inspected argument {} is not within function args.'
                        ' The function was: {}.'
                        .format(err_subject, tag_arg_name, function.__name__))

                tag_arg_val = call_args[tag_arg_name]

                # First, check that we have some value
                if not tag_arg_val:
                    raise ValidationError('{} Can not validate {}. No value provided.'
                                          .format(err_subject, tag_arg_name))

                # If any of them are a dict, then use the provided dict_key
                if isinstance(tag_arg_val, dict):
                    if dict_key not in tag_arg_val:
                        raise ProgrammingError(
                            '{} Inspected dict arg {} does not contain {} key.'
                            ' The function was: {}.'
                            .format(err_subject, tag_arg_name, dict_key, function.__name__))
                    tag_list = [tag_arg_val[dict_key]]
                elif isinstance(tag_arg_val, list):
                    tag_list = tag_arg_val
                else:
                    tag_list = [tag_arg_val]

                # Check to make sure the provided values match our whitelist.
                for allowed_prefix in conf.koji_tag_prefixes:
                    if all([t.startswith(pre + allowed_prefix + post) for t in tag_list]):
                        break
                else:
                    # Only raise this error if the given tags don't start with
                    # *any* of our allowed prefixes.
                    raise ValidationError(
                        'Koji tag validation: {} does not satisfy any of allowed prefixes: {}'
                        .format(tag_list,
                                [pre + p + post for p in conf.koji_tag_prefixes]))

            # Finally.. after all that validation, call the original function
            # and return its value.
            return function(*args, **kwargs)

        # We're replacing the original function with our synthetic wrapper,
        # but dress it up to make it look more like the original function.
        wrapper.__name__ = function.__name__
        wrapper.__doc__ = function.__doc__
        return wrapper

    return validation_decorator


def get_rpm_release(module_build):
    """
    Generates the dist tag for the specified module
    :param module_build: a models.ModuleBuild object
    :return: a string of the module's dist tag
    """
    dist_str = '.'.join([module_build.name, module_build.stream, str(module_build.version),
                         str(module_build.context)]).encode('utf-8')
    dist_hash = hashlib.sha1(dist_str).hexdigest()[:8]

    # We need to share the same auto-incrementing index in dist tag between all MSE builds.
    # We can achieve that by using the lowest build ID of all the MSE siblings including
    # this module build.
    mse_build_ids = module_build.siblings + [module_build.id or 0]
    mse_build_ids.sort()
    index = mse_build_ids[0]
    return "{prefix}{index}+{dist_hash}".format(
        prefix=conf.default_dist_tag_prefix,
        index=index,
        dist_hash=dist_hash,
    )


def create_dogpile_key_generator_func(skip_first_n_args=0):
    """
    Creates dogpile key_generator function with additional features:

    - when models.ModuleBuild is an argument of method cached by dogpile-cache,
      the ModuleBuild.id is used as a key. Therefore it is possible to cache
      data per particular module build, while normally, it would be per
      ModuleBuild.__str__() output, which contains also batch and other data
      which changes during the build of a module.
    - it is able to skip first N arguments of a cached method. This is useful
      when the db.session is part of cached method call, and the caching should
      work no matter what session instance is passed to cached method argument.
    """
    def key_generator(namespace, fn):
        fname = fn.__name__

        def generate_key(*arg, **kwarg):
            key_template = fname + "_"
            for s in arg[skip_first_n_args:]:
                if type(s) == models.ModuleBuild:
                    key_template += str(s.id)
                else:
                    key_template += str(s) + "_"
            return key_template

        return generate_key
    return key_generator


def import_mmd(session, mmd):
    """
    Imports new module build defined by `mmd` to MBS database using `session`.
    If it already exists, it is updated.

    The ModuleBuild.koji_tag is set according to xmd['mbs]['koji_tag'].
    The ModuleBuild.state is set to "ready".
    The ModuleBuild.rebuild_strategy is set to "all".
    The ModuleBuild.owner is set to "mbs_import".

    TODO: The "context" is not stored directly in database. We only store
    build_context and runtime_context and compute context, but when importing
    the module, we have no idea what build_context or runtime_context is - we only
    know the resulting "context", but there is no way to store it into do DB.
    By now, we just ignore mmd.get_context() and use default 00000000 context instead.
    """
    mmd.set_context("00000000")
    name = mmd.get_name()
    stream = mmd.get_stream()
    version = str(mmd.get_version())
    context = mmd.get_context()

    # NSVC is used for logging purpose later.
    nsvc = ":".join([name, stream, version, context])

    # Get the koji_tag.
    xmd = mmd.get_xmd()
    if "mbs" in xmd.keys() and "koji_tag" in xmd["mbs"].keys():
        koji_tag = xmd["mbs"]["koji_tag"]
    else:
        log.warn("'koji_tag' is not set in xmd['mbs'] for module %s", nsvc)
        koji_tag = ""

    # Get the ModuleBuild from DB.
    build = models.ModuleBuild.get_build_from_nsvc(
        session, name, stream, version, context)
    if build:
        log.info("Updating existing module build %s.", nsvc)
    else:
        build = models.ModuleBuild()

    build.name = name
    build.stream = stream
    build.version = version
    build.koji_tag = koji_tag
    build.state = models.BUILD_STATES['ready']
    build.modulemd = mmd.dumps()
    build.context = context
    build.owner = "mbs_import"
    build.rebuild_strategy = 'all'
    build.time_submitted = datetime.utcnow()
    build.time_modified = datetime.utcnow()
    build.time_completed = datetime.utcnow()
    session.add(build)
    session.commit()
    log.info("Module %s imported", nsvc)



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import mock

from mock import patch

import module_build_service.messaging
import module_build_service.scheduler.handlers.repos
import module_build_service.models
from tests import reuse_component_init_data
from tests import conf, db

import koji


class TestTagTagged:

    def setup_method(self, test_method):
        reuse_component_init_data()

    @mock.patch('module_build_service.models.ModuleBuild.from_tag_change_event')
    def test_no_matching_module(self, from_tag_change_event):
        """ Test that when a tag msg hits us and we have no match,
        that we do nothing gracefully.
        """
        from_tag_change_event.return_value = None
        msg = module_build_service.messaging.KojiTagChange(
            'no matches for this...', '2016-some-nonexistent-build', "artifact")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)

    def test_no_matching_artifact(self):
        """ Test that when a tag msg hits us and we have no match,
        that we do nothing gracefully.
        """
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c-build',
            "artifact")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)

    @patch("module_build_service.builder.GenericBuilder.default_buildroot_groups",
           return_value={'build': [], 'srpm-build': []})
    @patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.get_session")
    @patch("module_build_service.builder.GenericBuilder.create_from_module")
    def test_newrepo(self, create_builder, koji_get_session, dbg):
        """
        Test that newRepo is called in the expected times.
        """
        koji_session = mock.MagicMock()
        koji_session.getTag = lambda tag_name: {'name': tag_name}
        koji_session.getTaskInfo.return_value = {'state': koji.TASK_STATES['CLOSED']}
        koji_session.newRepo.return_value = 123456
        koji_get_session.return_value = koji_session

        builder = mock.MagicMock()
        builder.koji_session = koji_session
        builder.buildroot_ready.return_value = False
        builder.module_build_tag = {
            "name": "module-testmodule-master-20170219191323-c40c156c-build"}
        create_builder.return_value = builder

        module_build = module_build_service.models.ModuleBuild.query.filter_by(id=3).one()

        # Set previous components as COMPLETE and tagged.
        module_build.batch = 1
        for c in module_build.up_to_current_batch():
            c.state = koji.BUILD_STATES["COMPLETE"]
            c.tagged = True
            c.tagged_in_final = True

        module_build.batch = 2
        for c in module_build.current_batch():
            c.state = koji.BUILD_STATES["COMPLETE"]
        db.session.commit()

        # Tag the first component to the buildroot.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c-build',
            "perl-Tangerine")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)
        # Tag the first component to the final tag.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c',
            "perl-Tangerine")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)

        # newRepo should not be called, because there are still components
        # to tag.
        assert not koji_session.newRepo.called

        # Tag the second component to the buildroot.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c-build',
            "perl-List-Compare")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)

        # newRepo should not be called, because the component has not been
        # tagged to final tag so far.
        assert not koji_session.newRepo.called

        # Tag the first component to the final tag.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c',
            "perl-List-Compare")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)

        # newRepo should be called now - all components have been tagged.
        koji_session.newRepo.assert_called_once_with(
            "module-testmodule-master-20170219191323-c40c156c-build")

        # Refresh our module_build object.
        db.session.expunge(module_build)
        module_build = module_build_service.models.ModuleBuild.query.filter_by(id=3).one()

        # newRepo task_id should be stored in database, so we can check its
        # status later in poller.
        assert module_build.new_repo_task_id == 123456

    @patch("module_build_service.builder.GenericBuilder.default_buildroot_groups",
           return_value={'build': [], 'srpm-build': []})
    @patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.get_session")
    @patch("module_build_service.builder.GenericBuilder.create_from_module")
    def test_newrepo_still_building_components(self, create_builder, koji_get_session, dbg):
        """
        Test that newRepo is called in the expected times.
        """
        koji_session = mock.MagicMock()
        koji_session.getTag = lambda tag_name: {'name': tag_name}
        koji_session.getTaskInfo.return_value = {'state': koji.TASK_STATES['CLOSED']}
        koji_session.newRepo.return_value = 123456
        koji_get_session.return_value = koji_session

        builder = mock.MagicMock()
        builder.koji_session = koji_session
        builder.buildroot_ready.return_value = False
        builder.module_build_tag = {
            "name": "module-testmodule-master-20170219191323-c40c156c-build"}
        create_builder.return_value = builder

        module_build = module_build_service.models.ModuleBuild.query.filter_by(id=3).one()
        module_build.batch = 2
        component = module_build_service.models.ComponentBuild.query\
            .filter_by(package='perl-Tangerine', module_id=module_build.id).one()
        component.state = koji.BUILD_STATES["BUILDING"]
        db.session.commit()

        # Tag the perl-List-Compare component to the buildroot.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c-build',
            "perl-Tangerine")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)
        # Tag the perl-List-Compare component to final tag.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c',
            "perl-Tangerine")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)

        # newRepo should not be called, because perl-List-Compare has not been
        # built yet.
        assert not koji_session.newRepo.called

    @patch("module_build_service.builder.GenericBuilder.default_buildroot_groups",
           return_value={'build': [], 'srpm-build': []})
    @patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.get_session")
    @patch("module_build_service.builder.GenericBuilder.create_from_module")
    def test_newrepo_failed_components(self, create_builder, koji_get_session, dbg):
        """
        Test that newRepo is called in the expected times.
        """
        koji_session = mock.MagicMock()
        koji_session.getTag = lambda tag_name: {'name': tag_name}
        koji_session.getTaskInfo.return_value = {'state': koji.TASK_STATES['CLOSED']}
        koji_session.newRepo.return_value = 123456
        koji_get_session.return_value = koji_session

        builder = mock.MagicMock()
        builder.koji_session = koji_session
        builder.buildroot_ready.return_value = False
        builder.module_build_tag = {
            "name": "module-testmodule-master-20170219191323-c40c156c-build"}
        create_builder.return_value = builder

        module_build = module_build_service.models.ModuleBuild.query.filter_by(id=3).one()

        # Set previous components as COMPLETE and tagged.
        module_build.batch = 1
        for c in module_build.up_to_current_batch():
            c.state = koji.BUILD_STATES["COMPLETE"]
            c.tagged = True
            c.tagged_in_final = True

        module_build.batch = 2
        component = module_build_service.models.ComponentBuild.query\
            .filter_by(package='perl-Tangerine', module_id=module_build.id).one()
        component.state = koji.BUILD_STATES["FAILED"]
        component = module_build_service.models.ComponentBuild.query\
            .filter_by(package='perl-List-Compare', module_id=module_build.id).one()
        component.state = koji.BUILD_STATES["COMPLETE"]
        db.session.commit()

        # Tag the perl-List-Compare component to the buildroot.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c-build',
            "perl-List-Compare")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)
        # Tag the perl-List-Compare component to final tag.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c',
            "perl-List-Compare")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)

        # newRepo should be called now - all successfully built
        # components have been tagged.
        koji_session.newRepo.assert_called_once_with(
            "module-testmodule-master-20170219191323-c40c156c-build")

        # Refresh our module_build object.
        db.session.expunge(module_build)
        module_build = module_build_service.models.ModuleBuild.query.filter_by(id=3).one()

        # newRepo task_id should be stored in database, so we can check its
        # status later in poller.
        assert module_build.new_repo_task_id == 123456

    @patch("module_build_service.builder.GenericBuilder.default_buildroot_groups",
           return_value={'build': [], 'srpm-build': []})
    @patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.get_session")
    @patch("module_build_service.builder.GenericBuilder.create_from_module")
    def test_newrepo_multiple_batches_tagged(
            self, create_builder, koji_get_session, dbg):
        """
        Test that newRepo is called just once and only when all components
        are tagged even if we tag components from the multiple batches in the
        same time.
        """
        koji_session = mock.MagicMock()
        koji_session.getTag = lambda tag_name: {'name': tag_name}
        koji_session.getTaskInfo.return_value = {'state': koji.TASK_STATES['CLOSED']}
        koji_session.newRepo.return_value = 123456
        koji_get_session.return_value = koji_session

        builder = mock.MagicMock()
        builder.koji_session = koji_session
        builder.buildroot_ready.return_value = False
        builder.module_build_tag = {
            "name": "module-testmodule-master-20170219191323-c40c156c-build"}
        create_builder.return_value = builder

        module_build = module_build_service.models.ModuleBuild.query.filter_by(id=3).one()
        module_build.batch = 2
        mbm = module_build_service.models.ComponentBuild.query.filter_by(
            module_id=3, package='module-build-macros').one()
        mbm.tagged = False
        db.session.add(mbm)
        for c in module_build.current_batch():
            c.state = koji.BUILD_STATES["COMPLETE"]
        db.session.commit()

        # Tag the first component to the buildroot.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c-build',
            "perl-Tangerine")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)
        # Tag the first component to the final tag.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c',
            "perl-Tangerine")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)

        # newRepo should not be called, because there are still components
        # to tag.
        assert not koji_session.newRepo.called

        # Tag the second component to the buildroot.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c-build',
            "perl-List-Compare")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)
        # Tag the second component to final tag.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c',
            "perl-List-Compare")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)

        # newRepo should not be called, because there are still components
        # to tag.
        assert not koji_session.newRepo.called

        # Tag the component from first batch to final tag.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c',
            "module-build-macros")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)
        # Tag the component from first batch to the buildroot.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c-build',
            "module-build-macros")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)

        # newRepo should be called now - all components have been tagged.
        koji_session.newRepo.assert_called_once_with(
            "module-testmodule-master-20170219191323-c40c156c-build")

        # Refresh our module_build object.
        db.session.expunge(module_build)
        module_build = module_build_service.models.ModuleBuild.query.filter_by(id=3).one()

        # newRepo task_id should be stored in database, so we can check its
        # status later in poller.
        assert module_build.new_repo_task_id == 123456

    @patch("module_build_service.builder.GenericBuilder.default_buildroot_groups",
           return_value={'build': [], 'srpm-build': []})
    @patch("module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.get_session")
    @patch("module_build_service.builder.GenericBuilder.create_from_module")
    def test_newrepo_build_time_only(
            self, create_builder, koji_get_session, dbg):
        """
        Test the component.build_time_only is respected in tag handler.
        """
        koji_session = mock.MagicMock()
        koji_session.getTag = lambda tag_name: {'name': tag_name}
        koji_session.getTaskInfo.return_value = {'state': koji.TASK_STATES['CLOSED']}
        koji_session.newRepo.return_value = 123456
        koji_get_session.return_value = koji_session

        builder = mock.MagicMock()
        builder.koji_session = koji_session
        builder.buildroot_ready.return_value = False
        builder.module_build_tag = {
            "name": "module-testmodule-master-20170219191323-c40c156c-build"}
        create_builder.return_value = builder

        module_build = module_build_service.models.ModuleBuild.query.filter_by(id=3).one()

        # Set previous components as COMPLETE and tagged.
        module_build.batch = 1
        for c in module_build.up_to_current_batch():
            c.state = koji.BUILD_STATES["COMPLETE"]
            c.tagged = True
            c.tagged_in_final = True

        module_build.batch = 2
        component = module_build_service.models.ComponentBuild.query\
            .filter_by(package='perl-Tangerine', module_id=module_build.id).one()
        component.state = koji.BUILD_STATES["COMPLETE"]
        component.build_time_only = True
        component.tagged = False
        component.tagged_in_final = False
        component = module_build_service.models.ComponentBuild.query\
            .filter_by(package='perl-List-Compare', module_id=module_build.id).one()
        component.state = koji.BUILD_STATES["COMPLETE"]
        db.session.commit()

        # Tag the perl-Tangerine component to the buildroot.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c-build',
            "perl-Tangerine")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)
        assert not koji_session.newRepo.called
        # Tag the perl-List-Compare component to the buildroot.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c-build',
            "perl-List-Compare")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)
        # Tag the perl-List-Compare component to final tag.
        msg = module_build_service.messaging.KojiTagChange(
            'id',
            'module-testmodule-master-20170219191323-c40c156c',
            "perl-List-Compare")
        module_build_service.scheduler.handlers.tags.tagged(
            config=conf, session=db.session, msg=msg)

        # newRepo should be called now - all successfully built
        # components have been tagged.
        koji_session.newRepo.assert_called_once_with(
            "module-testmodule-master-20170219191323-c40c156c-build")

        # Refresh our module_build object.
        db.session.expunge(module_build)
        module_build = module_build_service.models.ModuleBuild.query.filter_by(id=3).one()

        # newRepo task_id should be stored in database, so we can check its
        # status later in poller.
        assert module_build.new_repo_task_id == 123456



# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from datetime import datetime

import pytest

import module_build_service.utils
from module_build_service import models, glib, Modulemd
from module_build_service.errors import StreamAmbigous
from tests import (db, clean_database)


class TestUtilsModuleStreamExpansion:

    def setup_method(self, test_method):
        clean_database(False)

    def teardown_method(self, test_method):
        clean_database()

    def _make_module(self, nsvc, requires_list, build_requires_list):
        """
        Creates new models.ModuleBuild defined by `nsvc` string with requires
        and buildrequires set according to `requires_list` and `build_requires_list`.

        :param str nsvc: name:stream:version:context of a module.
        :param list_of_dicts requires_list: List of dictionaries defining the
            requires in the mmd requires field format.
        :param list_of_dicts build_requires_list: List of dictionaries defining the
            build_requires_list in the mmd build_requires_list field format.
        :rtype: ModuleBuild
        :return: New Module Build.
        """
        name, stream, version, context = nsvc.split(":")
        mmd = Modulemd.Module()
        mmd.set_mdversion(2)
        mmd.set_name(name)
        mmd.set_stream(stream)
        mmd.set_version(int(version))
        mmd.set_context(context)
        mmd.set_summary("foo")
        mmd.set_description("foo")
        licenses = Modulemd.SimpleSet()
        licenses.add("GPL")
        mmd.set_module_licenses(licenses)

        if not isinstance(requires_list, list):
            requires_list = [requires_list]
        if not isinstance(build_requires_list, list):
            build_requires_list = [build_requires_list]

        xmd = {
            "mbs": {
                "buildrequires": [],
                "requires": [],
                "commit": "ref_%s" % context,
                "mse": "true",
            }
        }
        deps_list = []
        for requires, build_requires in zip(requires_list, build_requires_list):
            deps = Modulemd.Dependencies()
            for req_name, req_streams in requires.items():
                deps.add_requires(req_name, req_streams)
            for req_name, req_streams in build_requires.items():
                deps.add_buildrequires(req_name, req_streams)
            deps_list.append(deps)
        mmd.set_dependencies(deps_list)
        mmd.set_xmd(glib.dict_values(xmd))

        module_build = module_build_service.models.ModuleBuild()
        module_build.name = name
        module_build.stream = stream
        module_build.version = version
        module_build.context = context
        module_build.state = models.BUILD_STATES['ready']
        module_build.scmurl = 'git://pkgs.stg.fedoraproject.org/modules/unused.git?#ff1ea79'
        module_build.batch = 1
        module_build.owner = 'Tom Brady'
        module_build.time_submitted = datetime(2017, 2, 15, 16, 8, 18)
        module_build.time_modified = datetime(2017, 2, 15, 16, 19, 35)
        module_build.rebuild_strategy = 'changed-and-after'
        module_build.build_context = context
        module_build.stream_build_context = context
        module_build.runtime_context = context
        module_build.modulemd = mmd.dumps()
        db.session.add(module_build)
        db.session.commit()

        return module_build

    def _get_mmds_required_by_module_recursively(self, module_build):
        """
        Convenience wrapper around get_mmds_required_by_module_recursively
        returning the list with nsvc strings of modules returned by this the wrapped
        method.
        """
        mmd = module_build.mmd()
        module_build_service.utils.expand_mse_streams(db.session, mmd)
        modules = module_build_service.utils.get_mmds_required_by_module_recursively(
            db.session, mmd)
        nsvcs = [":".join([m.get_name(), m.get_stream(), str(m.get_version()), m.get_context()])
                 for m in modules]
        return nsvcs

    def _generate_default_modules(self):
        """
        Generates gtk:1, gtk:2, foo:1 and foo:2 modules requiring the
        platform:f28 and platform:f29 modules.
        """
        self._make_module("gtk:1:0:c2", {"platform": ["f28"]}, {})
        self._make_module("gtk:1:0:c3", {"platform": ["f29"]}, {})
        self._make_module("gtk:2:0:c4", {"platform": ["f28"]}, {})
        self._make_module("gtk:2:0:c5", {"platform": ["f29"]}, {})
        self._make_module("foo:1:0:c2", {"platform": ["f28"]}, {})
        self._make_module("foo:1:0:c3", {"platform": ["f29"]}, {})
        self._make_module("foo:2:0:c4", {"platform": ["f28"]}, {})
        self._make_module("foo:2:0:c5", {"platform": ["f29"]}, {})
        self._make_module("platform:f28:0:c10", {}, {})
        self._make_module("platform:f29:0:c11", {}, {})
        self._make_module("app:1:0:c6", {"platform": ["f29"]}, {})

    def test_generate_expanded_mmds_context(self):
        self._generate_default_modules()
        module_build = self._make_module(
            "app:1:0:c1", {"gtk": ["1", "2"]}, {"gtk": ["1", "2"]})
        mmds = module_build_service.utils.generate_expanded_mmds(
            db.session, module_build.mmd())
        contexts = set([mmd.get_context() for mmd in mmds])
        assert set(['e1e005fb', 'ce132a1e']) == contexts

    @pytest.mark.parametrize(
        'requires,build_requires,stream_ambigous,expected_xmd,expected_buildrequires', [
            ({"gtk": ["1", "2"]}, {"gtk": ["1", "2"]}, True,
             set([
                 frozenset(['platform:f28:0:c10', 'gtk:2:0:c4']),
                 frozenset(['platform:f28:0:c10', 'gtk:1:0:c2'])
             ]),
             set([
                 frozenset(['gtk:1']),
                 frozenset(['gtk:2']),
             ])),

            ({"gtk": ["1"], "foo": ["1"]}, {"gtk": ["1"], "foo": ["1"]}, False,
             set([
                 frozenset(['foo:1:0:c2', 'gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['foo:1', 'gtk:1'])
             ])),

            ({"gtk": ["1"], "foo": ["1"]}, {"gtk": ["1"], "foo": ["1"], "platform": ["f28"]}, False,
             set([
                 frozenset(['foo:1:0:c2', 'gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['foo:1', 'gtk:1'])
             ])),

            ({"gtk": ["-2"], "foo": ["-2"]}, {"gtk": ["-2"], "foo": ["-2"]}, True,
             set([
                 frozenset(['foo:1:0:c2', 'gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['foo:1', 'gtk:1'])
             ])),

            ({"gtk": ["1"], "foo": ["1"]}, {"gtk": ["-1", "1"], "foo": ["-2", "1"]}, False,
             set([
                 frozenset(['foo:1:0:c2', 'gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['foo:1', 'gtk:1'])
             ])),

            ({"gtk": ["1"], "foo": ["1"]}, {"gtk": ["1"]}, False,
             set([
                 frozenset(['gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['gtk:1'])
             ])),

            ({"gtk": []}, {"gtk": ["1"]}, True,
             set([
                 frozenset(['gtk:1:0:c2', 'platform:f28:0:c10'])
             ]),
             set([
                 frozenset(['gtk:1'])
             ])),

            ({}, {"app": ["1"]}, False,
             set([
                 frozenset(['app:1:0:c6', 'platform:f29:0:c11'])
             ]),
             set([
                 frozenset([])
             ])),
        ])
    def test_generate_expanded_mmds_buildrequires(
            self, requires, build_requires, stream_ambigous, expected_xmd,
            expected_buildrequires):
        self._generate_default_modules()
        module_build = self._make_module("app:1:0:c1", requires, build_requires)

        # Check that generate_expanded_mmds raises an exception if stream is ambigous
        # and also that it does not raise an exception otherwise.
        if stream_ambigous:
            with pytest.raises(StreamAmbigous):
                module_build_service.utils.generate_expanded_mmds(
                    db.session, module_build.mmd(), raise_if_stream_ambigous=True)
        else:
            module_build_service.utils.generate_expanded_mmds(
                db.session, module_build.mmd(), raise_if_stream_ambigous=True)

        # Check that if stream is ambigous and we define the stream, it does not raise
        # an exception.
        if stream_ambigous:
            default_streams = {}
            for ns in list(expected_buildrequires)[0]:
                name, stream = ns.split(":")
                default_streams[name] = stream
            module_build_service.utils.generate_expanded_mmds(
                db.session, module_build.mmd(), raise_if_stream_ambigous=True,
                default_streams=default_streams)

        mmds = module_build_service.utils.generate_expanded_mmds(
            db.session, module_build.mmd())

        buildrequires_per_mmd_xmd = set()
        buildrequires_per_mmd_buildrequires = set()
        for mmd in mmds:
            xmd = glib.from_variant_dict(mmd.get_xmd())
            br_nsvcs = []
            for name, detail in xmd['mbs']['buildrequires'].items():
                br_nsvcs.append(":".join([
                    name, detail["stream"], detail["version"], detail["context"]]))
            buildrequires_per_mmd_xmd.add(frozenset(br_nsvcs))

            assert len(mmd.get_dependencies()) == 1

            buildrequires = set()
            dep = mmd.get_dependencies()[0]
            for req_name, req_streams in dep.get_buildrequires().items():
                for req_stream in req_streams.get():
                    buildrequires.add(":".join([req_name, req_stream]))
            buildrequires_per_mmd_buildrequires.add(frozenset(buildrequires))

        assert buildrequires_per_mmd_xmd == expected_xmd
        assert buildrequires_per_mmd_buildrequires == expected_buildrequires

    @pytest.mark.parametrize('requires,build_requires,expected', [
        ({"gtk": ["1", "2"]}, {"gtk": ["1", "2"]},
         set([
             frozenset(['gtk:1']),
             frozenset(['gtk:2']),
         ])),

        ({"gtk": ["1", "2"]}, {"gtk": ["1"]},
         set([
             frozenset(['gtk:1', 'gtk:2']),
         ])),

        ({"gtk": ["1"], "foo": ["1"]}, {"gtk": ["1"], "foo": ["1"]},
         set([
             frozenset(['foo:1', 'gtk:1']),
         ])),

        ({"gtk": ["-2"], "foo": ["-2"]}, {"gtk": ["-2"], "foo": ["-2"]},
         set([
             frozenset(['foo:1', 'gtk:1']),
         ])),

        ({"gtk": ["-1", "1"], "foo": ["-2", "1"]}, {"gtk": ["-1", "1"], "foo": ["-2", "1"]},
         set([
             frozenset(['foo:1', 'gtk:1']),
         ])),

        ({"gtk": [], "foo": []}, {"gtk": ["1"], "foo": ["1"]},
         set([
             frozenset([]),
         ])),

    ])
    def test_generate_expanded_mmds_requires(self, requires, build_requires, expected):
        self._generate_default_modules()
        module_build = self._make_module("app:1:0:c1", requires, build_requires)
        mmds = module_build_service.utils.generate_expanded_mmds(
            db.session, module_build.mmd())

        requires_per_mmd = set()
        for mmd in mmds:
            assert len(mmd.get_dependencies()) == 1
            mmd_requires = set()
            dep = mmd.get_dependencies()[0]
            for req_name, req_streams in dep.get_requires().items():
                for req_stream in req_streams.get():
                    mmd_requires.add(":".join([req_name, req_stream]))
            requires_per_mmd.add(frozenset(mmd_requires))

        assert requires_per_mmd == expected

    @pytest.mark.parametrize('requires,build_requires,expected', [
        ({}, {"gtk": ["1", "2"]},
         ['platform:f29:0:c11', 'gtk:2:0:c4', 'gtk:2:0:c5',
          'platform:f28:0:c10', 'gtk:1:0:c2', 'gtk:1:0:c3']),

        ({}, {"gtk": ["1"], "foo": ["1"]},
         ['platform:f28:0:c10', 'gtk:1:0:c2', 'gtk:1:0:c3',
          'foo:1:0:c2', 'foo:1:0:c3', 'platform:f29:0:c11']),

        ({}, {"gtk": ["1"], "foo": ["1"], "platform": ["f28"]},
         ['platform:f28:0:c10', 'gtk:1:0:c2', 'gtk:1:0:c3',
          'foo:1:0:c2', 'foo:1:0:c3', 'platform:f29:0:c11']),

        ([{}, {}], [{"gtk": ["1"], "foo": ["1"]}, {"gtk": ["2"], "foo": ["2"]}],
         ['foo:1:0:c2', 'foo:1:0:c3', 'foo:2:0:c4', 'foo:2:0:c5',
          'platform:f28:0:c10', 'platform:f29:0:c11', 'gtk:1:0:c2',
          'gtk:1:0:c3', 'gtk:2:0:c4', 'gtk:2:0:c5']),

        ({}, {"gtk": ["-2"], "foo": ["-2"]},
         ['foo:1:0:c2', 'foo:1:0:c3', 'platform:f29:0:c11',
          'platform:f28:0:c10', 'gtk:1:0:c2', 'gtk:1:0:c3']),

        ({}, {"gtk": ["-1", "1"], "foo": ["-2", "1"]},
         ['foo:1:0:c2', 'foo:1:0:c3', 'platform:f29:0:c11',
          'platform:f28:0:c10', 'gtk:1:0:c2', 'gtk:1:0:c3']),
    ])
    def test_get_required_modules_simple(self, requires, build_requires, expected):
        module_build = self._make_module("app:1:0:c1", requires, build_requires)
        self._generate_default_modules()
        nsvcs = self._get_mmds_required_by_module_recursively(module_build)
        assert set(nsvcs) == set(expected)

    def _generate_default_modules_recursion(self):
        """
        Generates the gtk:1 module requiring foo:1 module requiring bar:1
        and lorem:1 modules which require base:f29 module requiring
        platform:f29 module :).
        """
        self._make_module("gtk:1:0:c2", {"foo": ["unknown"]}, {})
        self._make_module("gtk:1:1:c2", {"foo": ["1"]}, {})
        self._make_module("foo:1:0:c2", {"bar": ["unknown"]}, {})
        self._make_module("foo:1:1:c2", {"bar": ["1"], "lorem": ["1"]}, {})
        self._make_module("bar:1:0:c2", {"base": ["unknown"]}, {})
        self._make_module("bar:1:1:c2", {"base": ["f29"]}, {})
        self._make_module("lorem:1:0:c2", {"base": ["unknown"]}, {})
        self._make_module("lorem:1:1:c2", {"base": ["f29"]}, {})
        self._make_module("base:f29:0:c3", {"platform": ["f29"]}, {})
        self._make_module("platform:f29:0:c11", {}, {})

    @pytest.mark.parametrize('requires,build_requires,expected', [
        ({}, {"gtk": ["1"]},
         ['foo:1:1:c2', 'base:f29:0:c3', 'platform:f29:0:c11',
          'bar:1:1:c2', 'gtk:1:1:c2', 'lorem:1:1:c2']),

        ({}, {"foo": ["1"]},
         ['foo:1:1:c2', 'base:f29:0:c3', 'platform:f29:0:c11',
          'bar:1:1:c2', 'lorem:1:1:c2']),
    ])
    def test_get_required_modules_recursion(self, requires, build_requires, expected):
        module_build = self._make_module("app:1:0:c1", requires, build_requires)
        self._generate_default_modules_recursion()
        nsvcs = self._get_mmds_required_by_module_recursively(module_build)
        assert set(nsvcs) == set(expected)



# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>

""" Handlers for module change events on the message bus. """

from module_build_service import conf, models, log, build_logs
import module_build_service.builder
import module_build_service.resolver
import module_build_service.utils
import module_build_service.messaging
from module_build_service.utils import (
    attempt_to_reuse_all_components,
    record_component_builds,
    get_rpm_release,
    generate_koji_tag)
from module_build_service.errors import UnprocessableEntity, Forbidden, ValidationError

from requests.exceptions import ConnectionError

import koji

import logging
import os
import time

logging.basicConfig(level=logging.DEBUG)


def get_artifact_from_srpm(srpm_path):
    return os.path.basename(srpm_path).replace(".src.rpm", "")


def failed(config, session, msg):
    """
    Called whenever a module enters the 'failed' state.

    We cancel all the remaining component builds of a module
    and stop the building.
    """

    build = models.ModuleBuild.from_module_event(session, msg)

    module_info = build.json()
    if module_info['state'] != msg.module_build_state:
        log.warn("Note that retrieved module state %r "
                 "doesn't match message module state %r" % (
                     module_info['state'], msg.module_build_state))
        # This is ok.. it's a race condition we can ignore.
        pass

    unbuilt_components = [
        c for c in build.component_builds
        if (c.state != koji.BUILD_STATES['COMPLETE'] and
            c.state != koji.BUILD_STATES["FAILED"])
    ]

    if build.koji_tag:
        builder = module_build_service.builder.GenericBuilder.create_from_module(
            session, build, config)

        if build.new_repo_task_id:
            builder.cancel_build(build.new_repo_task_id)

        for component in unbuilt_components:
            if component.task_id:
                builder.cancel_build(component.task_id)
            component.state = koji.BUILD_STATES['FAILED']
            component.state_reason = build.state_reason
            session.add(component)

        # Tell the external buildsystem to wrap up
        builder.finalize()
    else:
        # Do not overwrite state_reason set by Frontend if any.
        if not build.state_reason:
            reason = "Missing koji tag. Assuming previously failed module lookup."
            log.error(reason)
            build.transition(config, state="failed", state_reason=reason)
            session.commit()
            return

    # Don't transition it again if it's already been transitioned
    if build.state != models.BUILD_STATES["failed"]:
        build.transition(config, state="failed")

    session.commit()

    build_logs.stop(build)
    module_build_service.builder.GenericBuilder.clear_cache(build)


def done(config, session, msg):
    """Called whenever a module enters the 'done' state.

    We currently don't do anything useful, so moving to ready.
    Otherwise the done -> ready state should happen when all
    dependent modules were re-built, at least that's the current plan.
    """
    build = models.ModuleBuild.from_module_event(session, msg)
    module_info = build.json()
    if module_info['state'] != msg.module_build_state:
        log.warn("Note that retrieved module state %r "
                 "doesn't match message module state %r" % (
                     module_info['state'], msg.module_build_state))
        # This is ok.. it's a race condition we can ignore.
        pass

    builder = module_build_service.builder.GenericBuilder.create_from_module(
        session, build, config)

    # Tell the external buildsystem to wrap up (CG import, createrepo, etc.)
    builder.finalize()

    build.transition(config, state="ready")
    session.commit()

    build_logs.stop(build)
    module_build_service.builder.GenericBuilder.clear_cache(build)


def init(config, session, msg):
    """ Called whenever a module enters the 'init' state."""
    # Sleep for a few seconds to make sure the module in the database is committed
    # TODO: Remove this once messaging is implemented in SQLAlchemy hooks
    for i in range(3):
        build = models.ModuleBuild.from_module_event(session, msg)
        if build:
            break
        time.sleep(1)

    try:
        mmd = build.mmd()
        record_component_builds(mmd, build, session=session)
        build.modulemd = mmd.dumps()
        build.transition(conf, models.BUILD_STATES["wait"])
    # Catch custom exceptions that we can expose to the user
    except (UnprocessableEntity, Forbidden, ValidationError, RuntimeError) as e:
        # Rollback changes underway
        session.rollback()
        build.transition(conf, models.BUILD_STATES["failed"], state_reason=str(e))
    except Exception as e:
        log.exception(str(e))
        # Rollback changes underway
        session.rollback()
        msg = "An unknown error occurred while validating the modulemd"
        build.transition(conf, models.BUILD_STATES["failed"], state_reason=msg)
    finally:
        session.add(build)
        session.commit()


def wait(config, session, msg):
    """ Called whenever a module enters the 'wait' state.

    We transition to this state shortly after a modulebuild is first requested.

    All we do here is request preparation of the buildroot.
    The kicking off of individual component builds is handled elsewhere,
    in module_build_service.schedulers.handlers.repos.
    """

    # Wait for the db on the frontend to catch up to the message, otherwise the
    # xmd information won't be present when we need it.
    # See https://pagure.io/fm-orchestrator/issue/386
    @module_build_service.utils.retry(interval=10, timeout=120, wait_on=RuntimeError)
    def _get_build_containing_xmd_for_mbs():
        build = models.ModuleBuild.from_module_event(session, msg)
        if 'mbs' in build.mmd().get_xmd():
            return build
        session.expire(build)
        raise RuntimeError("{!r} doesn't contain xmd information for MBS."
                           .format(build))

    build = _get_build_containing_xmd_for_mbs()
    build_logs.start(build)

    log.info("Found build=%r from message" % build)
    log.info("%r", build.modulemd)

    if build.state != msg.module_build_state:
        log.warn("Note that retrieved module state %r "
                 "doesn't match message module state %r" % (
                     build.state, msg.module_build_state))
        # This is ok.. it's a race condition we can ignore.
        pass

    tag = None
    dependencies = []

    resolver = module_build_service.resolver.GenericResolver.create(config)

    @module_build_service.utils.retry(
        interval=10, timeout=120,
        wait_on=(ValueError, RuntimeError, ConnectionError))
    def _get_deps_and_tag():
        """
        Private method to get the dependencies and koji tag of a module we
        are going to build. We use private method here to allow "retry"
        on failure.
        """
        cg_build_koji_tag = conf.koji_cg_default_build_tag
        if conf.system not in ['koji', 'test']:
            # In case of non-koji backend, we want to get the dependencies
            # of the local module build based on ModuleMetadata, because the
            # local build is not stored in the external MBS and therefore we
            # cannot query it using the `query` as for Koji below.
            dependencies = resolver.get_module_build_dependencies(
                mmd=build.mmd(), strict=True).keys()

            # We also don't want to get the tag name from the MBS, but just
            # generate it locally instead.
            tag = '-'.join(['module', build.name, build.stream, build.version])
        else:
            # For Koji backend, query for the module we are going to
            # build to get the koji_tag and deps from it.
            nsvc = ':'.join([build.name, build.stream, build.version])
            log.info("Getting deps for %s" % (nsvc))
            deps_dict = resolver.get_module_build_dependencies(
                build.name, build.stream, build.version, build.context, strict=True)
            dependencies = set(deps_dict.keys())

            # Find out the name of Koji tag to which the module's Content
            # Generator build should be tagged once the build finishes.
            module_names_streams = {mmd.get_name(): mmd.get_stream()
                                    for mmd in deps_dict.values()}
            for base_module_name in conf.base_module_names:
                if base_module_name in module_names_streams:
                    cg_build_koji_tag = conf.koji_cg_build_tag_template.format(
                        module_names_streams[base_module_name])
                    break

            log.info('Getting tag for {0}'.format(nsvc))
            tag = generate_koji_tag(build.name, build.stream, build.version, build.context)

        return dependencies, tag, cg_build_koji_tag

    try:
        dependencies, tag, cg_build_koji_tag = _get_deps_and_tag()
    except ValueError:
        reason = "Failed to get module info from MBS. Max retries reached."
        log.exception(reason)
        build.transition(config, state="failed", state_reason=reason)
        session.commit()
        raise

    log.debug("Found tag=%s for module %r" % (tag, build))
    # Hang on to this information for later.  We need to know which build is
    # associated with which koji tag, so that when their repos are regenerated
    # in koji we can figure out which for which module build that event is
    # relevant.
    log.debug("Assigning koji tag=%s to module build" % tag)
    build.koji_tag = tag

    log.debug("Assigning Content Generator build koji tag=%s to module "
              "build", cg_build_koji_tag)
    build.cg_build_koji_tag = cg_build_koji_tag

    builder = module_build_service.builder.GenericBuilder.create_from_module(
        session, build, config)

    log.debug("Adding dependencies %s into buildroot for module %s" % (dependencies, ':'.join(
        [build.name, build.stream, build.version])))
    builder.buildroot_add_repos(dependencies)

    if not build.component_builds:
        log.info("There are no components in module %r, skipping build" % build)
        build.transition(config, state="build")
        session.add(build)
        session.commit()
        # Return a KojiRepoChange message so that the build can be transitioned to done
        # in the repos handler
        return [module_build_service.messaging.KojiRepoChange(
            'handlers.modules.wait: fake msg', builder.module_build_tag['name'])]

    # If all components in module build will be reused, we don't have to build
    # module-build-macros, because there won't be any build done.
    if attempt_to_reuse_all_components(builder, session, build):
        log.info("All components have been reused for module %r, "
                 "skipping build" % build)
        build.transition(config, state="build")
        session.add(build)
        session.commit()
        return []

    log.debug("Starting build batch 1")
    build.batch = 1
    session.commit()

    artifact_name = "module-build-macros"

    component_build = models.ComponentBuild.from_component_name(
        session, artifact_name, build.id)
    further_work = []
    srpm = builder.get_disttag_srpm(
        disttag=".%s" % get_rpm_release(build),
        module_build=build)
    if not component_build:
        component_build = models.ComponentBuild(
            module_id=build.id,
            package=artifact_name,
            format="rpms",
            scmurl=srpm,
            batch=1,
            build_time_only=True
        )
        session.add(component_build)
        # Commit and refresh so that the SQLAlchemy relationships are available
        session.commit()
        session.refresh(component_build)
        msgs = builder.recover_orphaned_artifact(component_build)
        if msgs:
            log.info('Found an existing module-build-macros build')
            further_work += msgs
        # There was no existing artifact found, so lets submit the build instead
        else:
            task_id, state, reason, nvr = builder.build(artifact_name=artifact_name, source=srpm)
            component_build.task_id = task_id
            component_build.state = state
            component_build.reason = reason
            component_build.nvr = nvr
    elif component_build.state != koji.BUILD_STATES['COMPLETE']:
        # It's possible that the build succeeded in the builder but some other step failed which
        # caused module-build-macros to be marked as failed in MBS, so check to see if it exists
        # first
        msgs = builder.recover_orphaned_artifact(component_build)
        if msgs:
            log.info('Found an existing module-build-macros build')
            further_work += msgs
        else:
            task_id, state, reason, nvr = builder.build(artifact_name=artifact_name, source=srpm)
            component_build.task_id = task_id
            component_build.state = state
            component_build.reason = reason
            component_build.nvr = nvr

    session.add(component_build)
    build.transition(config, state="build")
    session.add(build)
    session.commit()

    # We always have to regenerate the repository.
    if config.system == "koji":
        log.info("Regenerating the repository")
        task_id = builder.koji_session.newRepo(
            builder.module_build_tag['name'])
        build.new_repo_task_id = task_id
        session.commit()
    else:
        further_work.append(module_build_service.messaging.KojiRepoChange(
            'fake msg', builder.module_build_tag['name']))
    return further_work



"""Sets the modulemd's context as the source of truth

Revision ID: 708ac8950f55
Revises: c8e2fc555399
Create Date: 2018-05-29 13:49:43.629831

"""

# revision identifiers, used by Alembic.
revision = '708ac8950f55'
down_revision = 'c8e2fc555399'

from alembic import op
import sqlalchemy as sa

# Data migration imports
from module_build_service import Modulemd
import hashlib


modulebuild = sa.Table(
    'module_builds',
    sa.MetaData(),
    sa.Column('id', sa.Integer, primary_key=True),
    sa.Column('build_context', sa.String()),
    sa.Column('runtime_context', sa.String()),
    sa.Column('modulemd', sa.String()),
    sa.Column('context', sa.String()),
)


def upgrade():
    connection = op.get_bind()

    for build in connection.execute(modulebuild.select()):
        if not build.modulemd:
            continue
        try:
            mmd = Modulemd.Module().new_from_string(build.modulemd)
            mmd.upgrade()
        except Exception:
            # If the modulemd isn't parseable then skip this build
            continue

        # If the context in the modulemd doesn't match the context in the DB,
        # then set the context in the DB to the one in the modulemd
        mmd_context = mmd.get_context()
        if mmd_context and mmd_context != build.context:
            connection.execute(
                modulebuild.update().where(modulebuild.c.id == build.id).values(
                    context=mmd_context))


def downgrade():
    connection = op.get_bind()

    for build in connection.execute(modulebuild.select()):
        if build.build_context and build.runtime_context:
            combined_hashes = '{0}:{1}'.format(
                build.build_context, build.runtime_context).encode('utf-8')
            context = hashlib.sha1(combined_hashes).hexdigest()[:8]
            connection.execute(
                modulebuild.update().where(modulebuild.c.id == build.id).values(context=context))



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>

import os

from tests.test_models import init_data, module_build_from_modulemd
from tests import init_data as init_data_contexts, clean_database
from module_build_service import conf, Modulemd
from module_build_service.models import ComponentBuild, ModuleBuild, make_session


class TestModels:
    def setup_method(self, test_method):
        init_data()

    def test_app_sqlalchemy_events(self):
        with make_session(conf) as session:
            component_build = ComponentBuild()
            component_build.package = 'before_models_committed'
            component_build.scmurl = \
                ('git://pkgs.domain.local/rpms/before_models_committed?'
                 '#9999999999999999999999999999999999999999')
            component_build.format = 'rpms'
            component_build.task_id = 999999999
            component_build.state = 1
            component_build.nvr = ('before_models_committed-0.0.0-0'
                                   '.module_before_models_committed_0_0')
            component_build.batch = 1
            component_build.module_id = 1

            session.add(component_build)
            session.commit()

        with make_session(conf) as session:
            c = session.query(ComponentBuild).filter(ComponentBuild.id == 1).one()
            assert c.component_builds_trace[0].id == 1
            assert c.component_builds_trace[0].component_id == 1
            assert c.component_builds_trace[0].state == 1
            assert c.component_builds_trace[0].state_reason is None
            assert c.component_builds_trace[0].task_id == 999999999

    def test_context_functions(self):
        """ Test that the build_context, runtime_context, and context hashes are correctly
        determined"""
        build = ModuleBuild.query.filter_by(id=1).one()
        yaml_path = os.path.join(
            os.path.dirname(__file__), '..', 'staged_data', 'testmodule_dependencies.yaml')
        mmd = Modulemd.Module.new_from_file(yaml_path)
        mmd.upgrade()
        build.modulemd = mmd.dumps()
        (build.ref_build_context, build.build_context, build.runtime_context,
         build.context) = ModuleBuild.contexts_from_mmd(build.modulemd)
        assert build.ref_build_context == 'f6e2aeec7576196241b9afa0b6b22acf2b6873d7'
        assert build.build_context == '089df24993c037e10174f3fa7342ab4dc191a4d4'
        assert build.runtime_context == 'bbc84c7b817ab3dd54916c0bcd6c6bdf512f7f9c'
        assert build.context == '3ee22b28'

    def test_siblings_property(self):
        """ Tests that the siblings property returns the ID of all modules with
        the same name:stream:version
        """
        clean_database()
        yaml_path = os.path.join(
            os.path.dirname(__file__), '..', 'staged_data', 'formatted_testmodule.yaml')
        mmd = Modulemd.Module.new_from_file(yaml_path)
        mmd.upgrade()
        with make_session(conf) as session:
            for i in range(3):
                build = module_build_from_modulemd(mmd.dumps())
                build.build_context = 'f6e2aeec7576196241b9afa0b6b22acf2b6873d' + str(i)
                build.runtime_context = 'bbc84c7b817ab3dd54916c0bcd6c6bdf512f7f9c' + str(i)
                session.add(build)
        session.commit()
        build_one = ModuleBuild.query.get(2)
        assert build_one.siblings == [3, 4]


class TestModelsGetStreamsContexts:

    def test_get_last_build_in_all_streams(self):
        init_data_contexts(contexts=True)
        with make_session(conf) as session:
            builds = ModuleBuild.get_last_build_in_all_streams(
                session, "nginx")
            builds = ["%s:%s:%s" % (build.name, build.stream, str(build.version))
                      for build in builds]
            assert builds == ["nginx:%d:%d" % (i, i + 2) for i in range(10)]

    def test_get_last_build_in_all_stream_last_version(self):
        init_data_contexts(contexts=False)
        with make_session(conf) as session:
            builds = ModuleBuild.get_last_build_in_all_streams(
                session, "nginx")
            builds = ["%s:%s:%s" % (build.name, build.stream, str(build.version))
                      for build in builds]
            assert builds == ["nginx:1:11"]

    def test_get_last_builds_in_stream(self):
        init_data_contexts(contexts=True)
        with make_session(conf) as session:
            builds = ModuleBuild.get_last_builds_in_stream(
                session, "nginx", "1")
            builds = ["%s:%s:%s:%s" % (build.name, build.stream, str(build.version),
                                       build.context) for build in builds]
            assert builds == ['nginx:1:3:d5a6c0fa', 'nginx:1:3:795e97c1']



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>
#            Matt Prahl <mprahl@redhat.com>
#            Jan Kaluza <jkaluza@redhat.com>
import re
import time
import shutil
import tempfile
import os
from multiprocessing.dummy import Pool as ThreadPool
from datetime import datetime

from module_build_service import conf, db, log, models, Modulemd
from module_build_service.errors import (
    ValidationError, UnprocessableEntity, Forbidden, Conflict)
from module_build_service import glib
import module_build_service.scm
from .mse import generate_expanded_mmds


def _scm_get_latest(pkg):
    try:
        # If the modulemd specifies that the 'f25' branch is what
        # we want to pull from, we need to resolve that f25 branch
        # to the specific commit available at the time of
        # submission (now).
        pkgref = module_build_service.scm.SCM(
            pkg.get_repository()).get_latest(pkg.get_ref())
    except Exception as e:
        log.exception(e)
        return {'error': "Failed to get the latest commit for %s#%s" % (
            pkg.get_repository(), pkg.get_ref())}

    return {
        'pkg_name': pkg.get_name(),
        'pkg_ref': pkgref,
        'error': None
    }


def format_mmd(mmd, scmurl, session=None):
    """
    Prepares the modulemd for the MBS. This does things such as replacing the
    branches of components with commit hashes and adding metadata in the xmd
    dictionary.
    :param mmd: the ModuleMetadata object to format
    :param scmurl: the url to the modulemd
    """
    # Import it here, because SCM uses utils methods and fails to import
    # them because of dep-chain.
    from module_build_service.scm import SCM

    if not session:
        session = db.session

    xmd = glib.from_variant_dict(mmd.get_xmd())
    if 'mbs' not in xmd:
        xmd['mbs'] = {}
    if 'scmurl' not in xmd['mbs']:
        xmd['mbs']['scmurl'] = scmurl or ''
    if 'commit' not in xmd['mbs']:
        xmd['mbs']['commit'] = ''

    local_modules = models.ModuleBuild.local_modules(session)
    local_modules = {m.name + "-" + m.stream: m for m in local_modules}

    # If module build was submitted via yaml file, there is no scmurl
    if scmurl:
        scm = SCM(scmurl)
        # If a commit hash is provided, add that information to the modulemd
        if scm.commit:
            # We want to make sure we have the full commit hash for consistency
            if SCM.is_full_commit_hash(scm.scheme, scm.commit):
                full_scm_hash = scm.commit
            else:
                full_scm_hash = scm.get_full_commit_hash()

            xmd['mbs']['commit'] = full_scm_hash
        # If a commit hash wasn't provided then just get the latest from master
        else:
            xmd['mbs']['commit'] = scm.get_latest()

    if mmd.get_rpm_components() or mmd.get_module_components():
        if 'rpms' not in xmd['mbs']:
            xmd['mbs']['rpms'] = {}
        # Add missing data in RPM components
        for pkgname, pkg in mmd.get_rpm_components().items():
            if pkg.get_repository() and not conf.rpms_allow_repository:
                raise Forbidden(
                    "Custom component repositories aren't allowed.  "
                    "%r bears repository %r" % (pkgname, pkg.get_repository()))
            if pkg.get_cache() and not conf.rpms_allow_cache:
                raise Forbidden(
                    "Custom component caches aren't allowed.  "
                    "%r bears cache %r" % (pkgname, pkg.cache))
            if not pkg.get_repository():
                pkg.set_repository(conf.rpms_default_repository + pkgname)
            if not pkg.get_cache():
                pkg.set_cache(conf.rpms_default_cache + pkgname)
            if not pkg.get_ref():
                pkg.set_ref('master')

        # Add missing data in included modules components
        for modname, mod in mmd.get_module_components().items():
            if mod.get_repository() and not conf.modules_allow_repository:
                raise Forbidden(
                    "Custom module repositories aren't allowed.  "
                    "%r bears repository %r" % (modname, mod.get_repository()))
            if not mod.get_repository():
                mod.set_repository(conf.modules_default_repository + modname)
            if not mod.get_ref():
                mod.set_ref('master')

        # Check that SCM URL is valid and replace potential branches in pkg refs
        # by real SCM hash and store the result to our private xmd place in modulemd.
        pool = ThreadPool(20)
        pkg_dicts = pool.map(_scm_get_latest, mmd.get_rpm_components().values())
        err_msg = ""
        for pkg_dict in pkg_dicts:
            if pkg_dict["error"]:
                err_msg += pkg_dict["error"] + "\n"
            else:
                pkg_name = pkg_dict["pkg_name"]
                pkg_ref = pkg_dict["pkg_ref"]
                xmd['mbs']['rpms'][pkg_name] = {'ref': pkg_ref}
        if err_msg:
            raise UnprocessableEntity(err_msg)

    # Set the modified xmd back to the modulemd
    mmd.set_xmd(glib.dict_values(xmd))


def validate_mmd(mmd):
    for modname, mod in mmd.get_module_components().items():
        if mod.get_repository() and not conf.modules_allow_repository:
            raise Forbidden(
                "Custom module repositories aren't allowed.  "
                "%r bears repository %r" % (modname, mod.get_repository()))


def merge_included_mmd(mmd, included_mmd):
    """
    Merges two modulemds. This merges only metadata which are needed in
    the `main` when it includes another module defined by `included_mmd`
    """
    included_xmd = glib.from_variant_dict(included_mmd.get_xmd())
    if 'rpms' in included_xmd['mbs']:
        xmd = glib.from_variant_dict(mmd.get_xmd())
        if 'rpms' not in xmd['mbs']:
            xmd['mbs']['rpms'] = included_xmd['mbs']['rpms']
        else:
            xmd['mbs']['rpms'].update(included_xmd['mbs']['rpms'])
    # Set the modified xmd back to the modulemd
    mmd.set_xmd(glib.dict_values(xmd))


def record_component_builds(mmd, module, initial_batch=1,
                            previous_buildorder=None, main_mmd=None, session=None):
    # Imported here to allow import of utils in GenericBuilder.
    import module_build_service.builder

    if not session:
        session = db.session

    # Format the modulemd by putting in defaults and replacing streams that
    # are branches with commit hashes
    format_mmd(mmd, module.scmurl, session=session)

    # When main_mmd is set, merge the metadata from this mmd to main_mmd,
    # otherwise our current mmd is main_mmd.
    if main_mmd:
        # Check for components that are in both MMDs before merging since MBS
        # currently can't handle that situation.
        duplicate_components = [rpm for rpm in main_mmd.get_rpm_components().keys()
                                if rpm in mmd.get_rpm_components()]
        if duplicate_components:
            error_msg = (
                'The included module "{0}" in "{1}" have the following '
                'conflicting components: {2}'.format(
                    mmd.get_name(), main_mmd.get_name(), ', '.join(duplicate_components)))
            raise UnprocessableEntity(error_msg)
        merge_included_mmd(main_mmd, mmd)
    else:
        main_mmd = mmd

    # If the modulemd yaml specifies components, then submit them for build
    rpm_components = mmd.get_rpm_components().values()
    module_components = mmd.get_module_components().values()
    all_components = list(rpm_components) + list(module_components)
    if not all_components:
        return

    rpm_weights = module_build_service.builder.GenericBuilder.get_build_weights(
        [c.get_name() for c in rpm_components])
    all_components.sort(key=lambda x: x.get_buildorder())
    # We do not start with batch = 0 here, because the first batch is
    # reserved for module-build-macros. First real components must be
    # planned for batch 2 and following.
    batch = initial_batch

    for component in all_components:
        # Increment the batch number when buildorder increases.
        if previous_buildorder != component.get_buildorder():
            previous_buildorder = component.get_buildorder()
            batch += 1

        # If the component is another module, we fetch its modulemd file
        # and record its components recursively with the initial_batch
        # set to our current batch, so the components of this module
        # are built in the right global order.
        if isinstance(component, Modulemd.ComponentModule):
            full_url = component.get_repository() + "?#" + component.get_ref()
            # It is OK to whitelist all URLs here, because the validity
            # of every URL have been already checked in format_mmd(...).
            included_mmd = _fetch_mmd(full_url, whitelist_url=True)[0]
            batch = record_component_builds(included_mmd, module, batch,
                                            previous_buildorder, main_mmd, session=session)
            continue

        component_ref = mmd.get_xmd()['mbs']['rpms'][component.get_name()]['ref']
        full_url = component.get_repository() + "?#" + component_ref
        build = models.ComponentBuild(
            module_id=module.id,
            package=component.get_name(),
            format="rpms",
            scmurl=full_url,
            batch=batch,
            ref=component_ref,
            weight=rpm_weights[component.get_name()]
        )
        session.add(build)

    return batch


def submit_module_build_from_yaml(username, handle, stream=None, skiptests=False,
                                  optional_params=None):
    yaml_file = handle.read()
    mmd = load_mmd(yaml_file)

    # Mimic the way how default values are generated for modules that are stored in SCM
    # We can take filename as the module name as opposed to repo name,
    # and also we can take numeric representation of current datetime
    # as opposed to datetime of the last commit
    dt = datetime.utcfromtimestamp(int(time.time()))
    def_name = str(os.path.splitext(os.path.basename(handle.filename))[0])
    def_version = int(dt.strftime("%Y%m%d%H%M%S"))
    mmd.set_name(mmd.get_name() or def_name)
    mmd.set_stream(stream or mmd.get_stream() or "master")
    mmd.set_version(mmd.get_version() or def_version)
    if skiptests:
        buildopts = mmd.get_rpm_buildopts()
        buildopts["macros"] = buildopts.get("macros", "") + "\n\n%__spec_check_pre exit 0\n"
        mmd.set_rpm_buildopts(buildopts)
    return submit_module_build(username, None, mmd, None, optional_params)


_url_check_re = re.compile(r"^[^:/]+:.*$")


def submit_module_build_from_scm(username, url, branch, allow_local_url=False,
                                 optional_params=None):
    # Translate local paths into file:// URL
    if allow_local_url and not _url_check_re.match(url):
        log.info(
            "'{}' is not a valid URL, assuming local path".format(url))
        url = os.path.abspath(url)
        url = "file://" + url
    mmd, scm = _fetch_mmd(url, branch, allow_local_url)

    return submit_module_build(username, url, mmd, scm, optional_params)


def submit_module_build(username, url, mmd, scm, optional_params=None):
    """
    Submits new module build.

    :param str username: Username of the build's owner.
    :param str url: SCM URL of submitted build.
    :param Modulemd.Module mmd: Modulemd defining the build.
    :param scm.SCM scm: SCM class representing the cloned git repo.
    :param dict optional_params: Dict with optional params for a build:
        - "local_build" (bool): The module is being built locally (the MBS is
          not running in infra, but on local developer's machine).
        - "default_streams" (dict): Dict with name:stream mapping defining the stream
          to choose for given module name if multiple streams are available to choose from.
        - Any optional ModuleBuild class field (str).
    :rtype: list with ModuleBuild
    :return: List with submitted module builds.
    """
    import koji  # Placed here to avoid py2/py3 conflicts...

    # For local builds, we want the user to choose the exact stream using the default_streams
    # in case there are multiple streams to choose from and raise an exception otherwise.
    if optional_params and "local_build" in optional_params:
        raise_if_stream_ambigous = True
        del optional_params["local_build"]
    else:
        raise_if_stream_ambigous = False

    # Get the default_streams if set.
    if optional_params and "default_streams" in optional_params:
        default_streams = optional_params["default_streams"]
        del optional_params["default_streams"]
    else:
        default_streams = {}

    validate_mmd(mmd)
    mmds = generate_expanded_mmds(db.session, mmd, raise_if_stream_ambigous, default_streams)
    modules = []

    for mmd in mmds:
        log.debug('Checking whether module build already exists: %s.',
                  ":".join([mmd.get_name(), mmd.get_stream(),
                           str(mmd.get_version()), mmd.get_context()]))
        module = models.ModuleBuild.get_build_from_nsvc(
            db.session, mmd.get_name(), mmd.get_stream(), str(mmd.get_version()),
            mmd.get_context())
        if module:
            if module.state != models.BUILD_STATES['failed']:
                err_msg = ('Module (state=%s) already exists. Only a new build or resubmission of '
                           'a failed build is allowed.' % module.state)
                log.error(err_msg)
                raise Conflict(err_msg)
            if optional_params:
                rebuild_strategy = optional_params.get('rebuild_strategy')
                if rebuild_strategy and module.rebuild_strategy != rebuild_strategy:
                    raise ValidationError(
                        'You cannot change the module\'s "rebuild_strategy" when '
                        'resuming a module build')
            log.debug('Resuming existing module build %r' % module)
            # Reset all component builds that didn't complete
            for component in module.component_builds:
                if component.state and component.state != koji.BUILD_STATES['COMPLETE']:
                    component.state = None
                    component.state_reason = None
                    db.session.add(component)
            module.username = username
            prev_state = module.previous_non_failed_state
            if prev_state == models.BUILD_STATES['init']:
                transition_to = models.BUILD_STATES['init']
            else:
                transition_to = models.BUILD_STATES['wait']
                module.batch = 0
            module.transition(conf, transition_to, "Resubmitted by %s" % username)
            log.info("Resumed existing module build in previous state %s" % module.state)
        else:
            log.debug('Creating new module build')
            module = models.ModuleBuild.create(
                db.session,
                conf,
                name=mmd.get_name(),
                stream=mmd.get_stream(),
                version=str(mmd.get_version()),
                modulemd=mmd.dumps(),
                scmurl=url,
                username=username,
                **(optional_params or {})
            )
            (module.ref_build_context, module.build_context, module.runtime_context,
             module.context) = module.contexts_from_mmd(module.modulemd)

        db.session.add(module)
        db.session.commit()
        modules.append(module)
        log.info("%s submitted build of %s, stream=%s, version=%s, context=%s", username,
                 mmd.get_name(), mmd.get_stream(), mmd.get_version(), mmd.get_context())
    return modules


def _fetch_mmd(url, branch=None, allow_local_url=False, whitelist_url=False):
    # Import it here, because SCM uses utils methods
    # and fails to import them because of dep-chain.
    import module_build_service.scm

    td = None
    scm = None
    try:
        log.debug('Verifying modulemd')
        td = tempfile.mkdtemp()
        if whitelist_url:
            scm = module_build_service.scm.SCM(url, branch, [url], allow_local_url)
        else:
            scm = module_build_service.scm.SCM(url, branch, conf.scmurls, allow_local_url)
        scm.checkout(td)
        scm.verify()
        cofn = scm.get_module_yaml()
        mmd = load_mmd(cofn, is_file=True)
    finally:
        try:
            if td is not None:
                shutil.rmtree(td)
        except Exception as e:
            log.warning(
                "Failed to remove temporary directory {!r}: {}".format(
                    td, str(e)))

    # If the name was set in the modulemd, make sure it matches what the scmurl
    # says it should be
    if mmd.get_name() and mmd.get_name() != scm.name:
        raise ValidationError('The name "{0}" that is stored in the modulemd '
                              'is not valid'.format(mmd.get_name()))
    else:
        mmd.set_name(scm.name)

    # If the stream was set in the modulemd, make sure it matches what the repo
    # branch is
    if mmd.get_stream() and mmd.get_stream() != scm.branch:
        raise ValidationError('The stream "{0}" that is stored in the modulemd '
                              'does not match the branch "{1}"'.format(
                                  mmd.get_stream(), scm.branch))
    else:
        mmd.set_stream(str(scm.branch))

    # If the version is in the modulemd, throw an exception since the version
    # since the version is generated by MBS
    if mmd.get_version():
        raise ValidationError('The version "{0}" is already defined in the '
                              'modulemd but it shouldn\'t be since the version '
                              'is generated based on the commit time'.format(
                                  mmd.get_version()))
    else:
        mmd.set_version(int(scm.version))

    return mmd, scm


def load_mmd(yaml, is_file=False):
    try:
        if is_file:
            mmd = Modulemd.Module().new_from_file(yaml)
        else:
            mmd = Modulemd.Module().new_from_string(yaml)
        # If the modulemd was v1, it will be upgraded to v2
        mmd.upgrade()
    except Exception:
        error = 'The following invalid modulemd was encountered: {0}'.format(yaml)
        log.exception(error)
        raise UnprocessableEntity(error)

    return mmd


def load_local_builds(local_build_nsvs, session=None):
    """
    Loads previously finished local module builds from conf.mock_resultsdir
    and imports them to database.

    :param local_build_nsvs: List of NSV separated by ':' defining the modules
        to load from the mock_resultsdir.
    """
    if not local_build_nsvs:
        return

    if not session:
        session = db.session

    if type(local_build_nsvs) != list:
        local_build_nsvs = [local_build_nsvs]

    # Get the list of all available local module builds.
    builds = []
    try:
        for d in os.listdir(conf.mock_resultsdir):
            m = re.match('^module-(.*)-([^-]*)-([0-9]+)$', d)
            if m:
                builds.append((m.group(1), m.group(2), int(m.group(3)), d))
    except OSError:
        pass

    # Sort with the biggest version first
    try:
        # py27
        builds.sort(lambda a, b: -cmp(a[2], b[2]))  # noqa: F821
    except TypeError:
        # py3
        builds.sort(key=lambda a: a[2], reverse=True)

    for nsv in local_build_nsvs:
        parts = nsv.split(':')
        if len(parts) < 1 or len(parts) > 3:
            raise RuntimeError(
                'The local build "{0}" couldn\'t be be parsed into '
                'NAME[:STREAM[:VERSION]]'.format(nsv))

        name = parts[0]
        stream = parts[1] if len(parts) > 1 else None
        version = int(parts[2]) if len(parts) > 2 else None

        found_build = None
        for build in builds:
            if name != build[0]:
                continue
            if stream is not None and stream != build[1]:
                continue
            if version is not None and version != build[2]:
                continue

            found_build = build
            break

        if not found_build:
            raise RuntimeError(
                'The local build "{0}" couldn\'t be found in "{1}"'.format(
                    nsv, conf.mock_resultsdir))

        # Load the modulemd metadata.
        path = os.path.join(conf.mock_resultsdir, found_build[3], 'results')
        mmd = load_mmd(os.path.join(path, 'modules.yaml'), is_file=True)

        # Create ModuleBuild in database.
        module = models.ModuleBuild.create(
            session,
            conf,
            name=mmd.get_name(),
            stream=mmd.get_stream(),
            version=str(mmd.get_version()),
            context=mmd.get_context(),
            modulemd=mmd.dumps(),
            scmurl="",
            username="mbs",
            publish_msg=False)
        module.koji_tag = path
        module.state = models.BUILD_STATES['ready']
        session.commit()

        if (found_build[0] != module.name or found_build[1] != module.stream or
                str(found_build[2]) != module.version):
            raise RuntimeError(
                'Parsed metadata results for "{0}" don\'t match the directory name'
                .format(found_build[3]))
        log.info("Loaded local module build %r", module)
